// Copyright 2004,2005 Luke Campagnola <lcampagn@mines.edu>, Colorado School of Mines
// 
// This file is part of VRManager.
// 
// VRManager is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
// 
// VRManager is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with VRManager; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include "Pointer.h"
#include "debug.h"
#include "xmlHelpers.h"
#include "vrmanager.h"

Pointer* createNewPointer(xmlNode* node, class VRManager* vrm, char* name) {
  Pointer* ret;
  // Set the pointer type if needed
  if( nodeHasProp(node, "alias") ) {
    xmlSetProp(node, "type", "alias");
  }
  else if( nodeHasProp(node, "trackerSystem") ) {
    xmlSetProp(node, "type", "tracked");
  }
  if( ! nodeHasProp(node, "type") ) {
    xmlSetProp(node, "type", "static");
  }
  
  
  // create the proper pointer
  if( xmlPropEq(node, "type", "static") ) {
    ret = new Pointer_Static(node, vrm);
  }
  else if( xmlPropEq(node, "type", "alias") ) {
    ret = new Pointer_Alias(node, vrm);
  }
  else if( xmlPropEq(node, "type", "tracked") ) {
    ret = new Pointer_Tracked(node, vrm);
  }
  else if( xmlPropEq(node, "type", "centroid") ) {
    ret = new Pointer_Centroid(node, vrm);
  }
  else if( xmlPropEq(node, "type", "intersection") ) {
    ret = new Pointer_Intersection(node, vrm);
  }
  else {
    msg(M_WARN, "Pointer type \"%s\" not supported.\n", xmlGetProp(node, "type"));
  }
  
  return ret;
}







  // base class
  
Pointer::Pointer() {

}


Pointer::Pointer(xmlNode* node, VRManager* vrm) {
  setVrm(vrm);
  setVrmName(xmlGetProp(node, "name"));
}





  // static pointer
  
Pointer_Static::Pointer_Static() {
}

Pointer_Static::Pointer_Static(xmlNode* node, VRManager* vrm) : Pointer(node, vrm) {
  reconfigure(node);
}

void Pointer_Static::reconfigure(xmlNode* node) {
  clearDeps();
  assertNodeHasProp(node, "vector");
  position = strToVector(xmlGetProp(node, "vector"));
}

Pointer_Static::Pointer_Static(Vector p) {
  position = p;
}

const Vector Pointer_Static::getLocation() {
  return(position);
}

const Matrix Pointer_Static::getTransform() {
  return(translateMatrix(position));
}

const Matrix Pointer_Static::getInvTransform() {
  return(translateMatrix(Vector() - position));
}


void Pointer_Static::setLocation(Vector newVector) {
  position = newVector;
  informUpdate();
}
  
 // Alias Pointer

Pointer_Alias::Pointer_Alias() {

}

Pointer_Alias::Pointer_Alias(xmlNode* node, VRManager* vrm) : Pointer(node, vrm) {
  reconfigure(node);
}

void Pointer_Alias::reconfigure(xmlNode* node) {
  clearDeps();
  assertNodeHasProp(node, "alias");
  alias = getVrm()->getPointer(xmlGetProp(node, "alias"));
  if( alias == NULL ) {
    msg(M_ERR, "Cannot make alias of non-existing pointer \"%s\"\n", xmlGetProp(node, "alias"));
  }
  else {
    alias->addDependent(this);
  }
}

const Vector Pointer_Alias::getLocation() {
  return(alias->getLocation());
}

const Matrix Pointer_Alias::getTransform() {
  return(alias->getTransform());
}

const Matrix Pointer_Alias::getInvTransform() {
  return(alias->getInvTransform());
}

/*void Pointer_Alias::setLocation(Vector newVector) {
  alias->setLocation(newVector);
}*/
 
 
 
 // Tracked pointer

void Pointer_Tracked::depChanged(int msg) {
  posValid = transValid = false;
}
  
Pointer_Tracked::Pointer_Tracked() {
  posValid = transValid = false;
}


Pointer_Tracked::Pointer_Tracked(xmlNode* node, VRManager* vrm) : Pointer(node, vrm) {
  reconfigure(node);
  tkr->addDependent(this);
}

void Pointer_Tracked::reconfigure(xmlNode* node) {
  clearDeps();
  posValid = transValid = false;
  assertNodeHasProp(node, "trackerSystem");
  assertNodeHasProp(node, "tracker");
  char* TSName = xmlGetProp(node, "trackerSystem");
  char* TName = xmlGetProp(node, "tracker");
  tkrPos = strToVector(xmlGetProp(node, "vector"));
  if( getVrm()->getTrackerSystem(TSName) == NULL ) {
    msg(M_ERR, "trackerSystem \"%s\" not found.\n", TSName);
  }
  tkr = getVrm()->getTrackerSystem(TSName)->getTracker(TName);
  if( tkr == NULL ) {
    msg(M_ERR, "tracker \"%s\" not found within trackerSystem \"%s\" (0x%lx).\n", TName, TSName, getVrm()->getTrackerSystem(TSName));
  }
}

Pointer_Tracked::Pointer_Tracked(Tracker* new_tkr, Vector pos) {
  posValid = transValid = false;
  tkr = new_tkr;
  tkrPos = pos;
  tkr->addDependent(this);
}

void Pointer_Tracked::setTracker(Tracker* t) {
  tkr = t;
}

void Pointer_Tracked::setPosition(Vector p) {
  tkrPos = p;
}


const Vector Pointer_Tracked::getLocation() {
  if( ! posValid ) {
    tkrTransform = tkr->getTransformMatrix();
    realPos = *tkrTransform * tkrPos;
    posValid = true;
  }
  return( realPos );
}

const Matrix Pointer_Tracked::getTransform() {
  if( ! transValid ) {
    tkrTransform = tkr->getTransformMatrix();
    realTransform = *tkrTransform * translateMatrix(tkrPos);
    transValid = true;
  }
    
  return( realTransform );
}

const Matrix Pointer_Tracked::getInvTransform() {
  if( ! invTransValid ) {
  
      // TODO: This can possibly be made faster by deriving the
      // inverse matrix directly from tracking information..
    realInvTransform = getTransform().inverse();
    invTransValid = true;
  }
    
  return( realInvTransform );
}


  // centroid pointer
  


void Pointer_Centroid::depChanged(int msg) {
  posValid = transValid = invTransValid = false;
}

Pointer_Centroid::Pointer_Centroid() {
  posValid = transValid = invTransValid = false;
}



Pointer_Centroid::Pointer_Centroid(xmlNode* node, VRManager* vrm) : Pointer(node, vrm) {
  reconfigure(node);
}


void Pointer_Centroid::reconfigure(xmlNode* node) {
  clearPointers();
  clearDeps();
  
  assertNodeHasProp(node, "pointers");

    // Add pointers now
/*  char* str = new char[strlen(xmlGetProp(node, "pointers"))+1];
  strncpy(str, xmlGetProp(node, "pointers"), strlen(xmlGetProp(node, "pointers")));*/
  char* str = strdup(xmlGetProp(node, "pointers"));
  char* tok;
  tok = strtok(str, ",");
  slist<char*> ptrs;
  while(tok != NULL) {
    while( tok[0] == ' ' )
      tok++;
    ptrs.push_front(tok);
    tok = strtok(NULL, ",");
  }
  
  for( slist<char*>::iterator i = ptrs.begin(); i != ptrs.end(); i++ ) {
    msg(M_INFO, "Adding pointer \"%s\" to centroid\n", *i);
    Pointer* p = getVrm()->getPointer(*i);
    if( p == NULL ) {
      msg(M_ERR, "Pointer \"%s\" does not exist; not adding to centroid\n", *i);
    }
    else {
      addPointer(p);
    }
  }
}



void Pointer_Centroid::addPointer(Pointer* ptr) {
  for( slist<Pointer*>::iterator i = pointers.begin(); i != NULL; i++ ) {
    if( ptr == *i ) {
      msg(M_INFO, "Not re-adding pointer \"%s\" to centroid.\n", ptr->getVrmName());
      return;
    }
  }  
  pointers.push_front(ptr);
  ptr->addDependent(this);
  informUpdate();
}

void Pointer_Centroid::clearPointers() {
  pointers.clear();
  informUpdate();
}

const Vector Pointer_Centroid::getLocation() {
  if( ! posValid ) {
    realPos.set();
    int num = 0;
    for( slist<Pointer*>::iterator i = pointers.begin(); i != NULL; i++ ) {
      realPos += (*i)->getLocation();
      num++;
    }
    realPos /= num;
    posValid = true;
  }
  
  return( realPos );
}

const Matrix Pointer_Centroid::getTransform() {
  if( ! transValid ) {
    Pointer* b = *(pointers.begin());
    realTransform = b->getTransform() * translateMatrix(b->getLocation() - getLocation());
    transValid = true;
  }
  return( realTransform );
}

const Matrix Pointer_Centroid::getInvTransform() {
  if( ! invTransValid ) {
    realInvTransform = getTransform().inverse();
    invTransValid = true;
  }
  return( realInvTransform );
}


  // intersection pointer
  
void Pointer_Intersection::depChanged(int msg) {
  posValid = transValid = invTransValid = false;
}

Pointer_Intersection::Pointer_Intersection() {
  posValid = transValid = invTransValid = false;
}

Pointer_Intersection::Pointer_Intersection(xmlNode* node, VRManager* vrm) : Pointer(node, vrm) {
  reconfigure(node);
}

void Pointer_Intersection::reconfigure(xmlNode* node) {
  clearDeps();
  assertNodeHasProp(node, "surface");
  assertNodeHasProp(node, "pointer1");
  assertNodeHasProp(node, "pointer2");
  Surface* sn = getVrm()->getSurface(xmlGetProp(node, "surface"));
  Pointer* an = getVrm()->getPointer(xmlGetProp(node, "pointer1"));
  Pointer* bn = getVrm()->getPointer(xmlGetProp(node, "pointer2"));

  if( sn == NULL ) {
    msg(M_ERR, "No surface named \"%s\"\n", xmlGetProp(node, "surface"));
  }
  if(an == NULL ) {
    msg(M_ERR, "No pointer named \"%s\"\n", xmlGetProp(node, "pointer1"));
  }
  if( bn == NULL ) {
    msg(M_ERR, "No pointer named \"%s\"\n", xmlGetProp(node, "pointer2"));
  }
  setObjects(an, bn, sn);
}

Pointer_Intersection::Pointer_Intersection(Pointer* a_new, Pointer* b_new, Surface* s_new) {
  posValid = transValid = invTransValid = false;
  setObjects(a_new, b_new, s_new);
}

void Pointer_Intersection::setObjects(Pointer* a_new, Pointer* b_new, Surface* s_new) {
  a = a_new;
  a->addDependent(this);
  b = b_new;
  b->addDependent(this);
  s = s_new;
  s->addDependent(this);
//  depChanged();
}

const Vector Pointer_Intersection::getLocation() {
  if( ! posValid ) {
  //  VRManager* v = getVrm();
  //  realPos.set(s->intersect(v->getPointerLocation(a), v->getPointerLocation(b)));
    realPos.set(s->intersect(a->getLocation(), b->getLocation()));
  }
  
  return( realPos );
}

const Matrix Pointer_Intersection::getTransform() {
  if( ! transValid ) {
    realTransform = translateMatrix(getLocation() - a->getLocation()) * a->getTransform();
    transValid = true;
  }
  return( realTransform );
}

const Matrix Pointer_Intersection::getInvTransform() {
  if( ! invTransValid ) {
    realInvTransform = getTransform().inverse();
    invTransValid = true;
  }
  return( realInvTransform );
}
