// Copyright 2004,2005 Luke Campagnola <lcampagn@mines.edu>, Colorado School of Mines
// 
// This file is part of VRManager.
// 
// VRManager is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
// 
// VRManager is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with VRManager; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#ifndef VRM_POINTER_H
#define VRM_POINTER_H

#include "vectormath.h"
#include "Tracker.h"
#include <slist.h>
#include "Comm.h"
#include "Surface.h"
#include "vrmObject.h"
#include "xmlHelpers.h"

  // Pointers represent any significant position/orientation in space, and can be referenced by the end user  

  
  
    
  // Pointer base class
class Pointer : public Comm, public vrmObject {
  public: 

    Pointer();
    Pointer(xmlNode* node, class VRManager* vrm);
    virtual void reconfigure(xmlNode* node) = 0;
    virtual const Vector getLocation() = 0;
    virtual const Matrix getTransform() = 0;
    virtual const Matrix getInvTransform() = 0;
//    virtual void setLocation(const Vector newVector) = 0;
  
};


  // Used by VRManager to create pointers from XML
Pointer* createNewPointer(xmlNode* node, class VRManager* vrm, char* name);

  // Static pointer;  never moves and has no orientation.  
class Pointer_Static : public Pointer {
  private:
    
    Vector position;
    Matrix transform;

  public: 

    Pointer_Static();
    Pointer_Static(xmlNode* node, VRManager* vrm);
    Pointer_Static(const Vector p);
    void reconfigure(xmlNode* node);
    const Vector getLocation();
    const Matrix getTransform();
    const Matrix getInvTransform();
    void setLocation(const Vector newVector);
  
};

  //////// Non-static pointer types ////////

class Pointer_Alias : public Pointer {
  private:
    Pointer* alias;
    
  public: 

    Pointer_Alias();
    Pointer_Alias(xmlNode* node, class VRManager* vrm);
    void reconfigure(xmlNode* node);
    const Vector getLocation();
    const Matrix getTransform();
    const Matrix getInvTransform();
//     void setLocation(const Vector newVector);
};
  
  
  // tracked pointer; has a position relative to a tracker and inherits tracker's orientation
class Pointer_Tracked : public Pointer {

  private:
  
    Vector position;
    Matrix transform;
    Tracker* tkr;
    Vector tkrPos;
    Matrix* tkrTransform;
    Matrix* tkrInvTransform;
    Vector realPos;
    Matrix realTransform;
    Matrix realInvTransform;
    bool posValid, transValid, invTransValid;
    void depChanged(int msg);

  public: 
  
    Pointer_Tracked();
    Pointer_Tracked(Tracker* tkr, Vector pos);
    Pointer_Tracked(xmlNode* node, class VRManager* vrm);
    void reconfigure(xmlNode* node);
    void setTracker(Tracker* t);
    void setPosition(Vector p);
    const Vector getLocation();
    const Matrix getTransform();
    const Matrix getInvTransform();

};

  // centroid pointer; Lives at the centroid of a group of pointers, inherits the first pointer's orientation.
class Pointer_Centroid : public Pointer {

  private:
  
    Vector position;
    Matrix transform;
    slist<Pointer*> pointers;
    Vector realPos;
    Matrix realTransform, realInvTransform;
    bool posValid, transValid, invTransValid;
    void depChanged(int msg=0);
  
  public: 
  
    Pointer_Centroid();
    Pointer_Centroid(xmlNode* node, class VRManager* vrm);
    void reconfigure(xmlNode* node);
    void addPointer(Pointer* ptr);
    void clearPointers();
    const Vector getLocation();
    const Matrix getTransform();
    const Matrix getInvTransform();

};
  
  
  
  // intersection pointer; Lives at the intersection of a line (defined by 2 pointers) and a Surface
class Pointer_Intersection : public Pointer {

  private:
  
    Vector position;
    Matrix transform;
    Pointer *a, *b;
    class Surface* s;
    
    Vector realPos;
    Matrix realTransform, realInvTransform;
    bool posValid, transValid, invTransValid;
    void depChanged(int msg=0);
  
  public: 
  
    Pointer_Intersection();
    Pointer_Intersection(xmlNode* node, VRManager* vrm);
    Pointer_Intersection(Pointer* a, Pointer* b, Surface* s);
    void reconfigure(xmlNode* node);
    void setObjects(Pointer* a, Pointer* b, Surface* s);
    const Vector getLocation();
    const Matrix getTransform();
    const Matrix getInvTransform();

};


#endif
