// Copyright 2004,2005 Luke Campagnola <lcampagn@mines.edu>, Colorado School of Mines
// 
// This file is part of VRManager.
// 
// VRManager is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
// 
// VRManager is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with VRManager; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include "Surface.h"
#include "debug.h"
#include "Viewport.h"
#include "Pointer.h"
#include "vrmanager.h"

Surface* createNewSurface(xmlNode* node, VRManager* vrm) {
  Surface* ret = NULL;
  // Set the pointer type if needed
  if( nodeHasProp(node, "alias") ) {
    xmlSetProp(node, "type", "alias");
  }
  if( ! nodeHasProp(node, "type") ) {
    xmlSetProp(node, "type", "3Point");
  }
  
    // If the surface is tracked, pass the info on to its pointers
  if( nodeHasProp(node, "trackerSystem") ) {
    assertNodeHasProp(node, "tracker");
    
    setChildrenProp(node, "pointer", "trackerSystem", xmlGetProp(node, "trackerSystem"));
    setChildrenProp(node, "pointer", "tracker", xmlGetProp(node, "tracker"));
  }
  
  // create the proper pointer
  if( xmlPropEq(node, "type", "alias") ) {
    ret = new Surface_Alias(node, vrm);
  }
  else if( xmlPropEq(node, "type", "3point") ) {
    ret = new Surface_3Point(node, vrm);
  }
  else if( xmlPropEq(node, "type", "2point") ) {
    ret = new Surface_2Point(node, vrm);
  }
  else {
    msg(M_WARN, "Surface type \"%s\" not supported.\n", xmlGetProp(node, "type"));
  }
  
  return ret;

}








Surface::Surface() {
  changed = false;
}

Surface::Surface(char* name, VRManager* vrm) {
  setVrm(vrm);
  setVrmName(name);
}

void Surface::setPlane(Vector o, Vector n) {
  origin.set(o);
  normal.set(n);
  d = - normal.dot(origin);
  changed = false;
}

void Surface::setPlane(Vector a, Vector b, Vector c) {
  setPlane(a, (c-a)*(b-a));
}

void Surface::update() {
}

bool Surface::isSimilar(Surface* a) {
  if( normal.angle(a->normal) < 0.05 ) {
    float dist = (intersect(a->origin, a->origin+a->normal)-a->origin).mag();
    if( dist < .05 ) {
      return true;
    }
    else {
      return false;
    }
  }
  else {
    return false;
  }
}

void Surface::addViewport(Viewport* v) {
  for( slist<Viewport*>::iterator i = viewports.begin(); i != NULL; i++ ) {
    if( v == (*i) ) {
      msg(M_INFO, "Surface: Not re-adding viewport \"%s\".\n", v->getVrmName());
      return;
    }
    else {
      viewports.push_front(v);
//      maskSet.addViewport(v);
    }
  }
}

// void Surface::setMasks() {
//   update();
//   for( slist<Viewport*>::iterator i = viewports.begin(); i != NULL; i++ ) {
// 
//   }
// }

Vector Surface::intersect(Vector a, Vector b) {
  update();
  Vector dx = b-a;
  float t = (-d - normal.dot(a)) / normal.dot(dx);
  return a + dx * t;  
}
 

void Surface::drawMask(Viewport* vp) {
//  maskSet.drawMask(vp->getVrmName());
}




Surface_Alias::Surface_Alias() {
}

Surface_Alias::Surface_Alias(xmlNode* node, VRManager* vrm) : Surface(xmlGetProp(node, "name"), vrm) {
  reconfigure(node);
}

bool Surface_Alias::isSimilar(Surface* s) {
  return alias->isSimilar(s);
}

void Surface_Alias::addViewport(Viewport* v) {
  alias->addViewport(v);
}

Vector Surface_Alias::intersect(Vector a, Vector b) {
  return alias->intersect(a, b);
}

void Surface_Alias::drawMask(Viewport* vp) {
  alias->drawMask(vp);
}

void Surface_Alias::reconfigure(xmlNode* node) {
  assertNodeHasProp(node, "alias");
  alias = getVrm()->getSurface(xmlGetProp(node, "alias"));
  alias->addDependent(this);
}








    
Surface_2Point::Surface_2Point() {
}

Surface_2Point::Surface_2Point(xmlNode* node, VRManager* vrm) : Surface(xmlGetProp(node, "name"), vrm) {
  reconfigure(node);
}

void Surface_2Point::reconfigure(xmlNode* node) {
  VRManager* vrm = getVrm();
  char* name = xmlGetProp(node, "name");
  xmlNode* cur;
  
  forEachChild(node, cur) {
    if( xmlNameEq(cur, "pointer") ) {
      // prefix pointer name with surface name
      char* p_name_orig = xmlGetProp(cur, "name");
      char* p_name = new char[strlen(name) + strlen(p_name_orig) + 2];
      sprintf(p_name, "%s_%s", name, p_name_orig);
      xmlSetProp(cur, "name", p_name);

      
      if( strcmp(p_name_orig, "o") == 0 )
        o = vrm->parsePointer(cur);
      else if( strcmp(p_name_orig, "n") == 0 )
        n = vrm->parsePointer(cur);
      else {
        msg(M_ERR, "Pointer named \"%s\" not allowed inside of <surface type=\"2point\">. Use o or n\n", xmlGetProp(cur, "name"));
      }
    }
    else {
      msg(M_WARN, "Only <pointer> tags allowed inside <surface>. Ignoring \"%s\"\n", xmlNodeName(cur));
    }
  }
}

Surface_2Point::Surface_2Point(Pointer* o_new, Pointer* n_new, VRManager* vrm, char* name) : Surface(name, vrm) {
  setPointers(o_new, n_new);
}

void Surface_2Point::setPointers(Pointer* o_new, Pointer* n_new) {
  o = o_new;
  o->addDependent(this);
  n = n_new;
  n->addDependent(this);
//  depChanged();
}

void Surface_2Point::depChanged(int msg) {
  bool changed = true;
}

void Surface_2Point::update() {
//  setPlane(getVrm()->getPointerLocation(o), getVrm()->getPointerLocation(n));
  setPlane(o->getLocation(), n->getLocation());
}


Surface_3Point::Surface_3Point() {
}

Surface_3Point::Surface_3Point(xmlNode* node, VRManager* vrm) : Surface(xmlGetProp(node, "name"), vrm) {
  reconfigure(node);
}

void Surface_3Point::reconfigure(xmlNode* node) {
  VRManager* vrm = getVrm();
  char* name = xmlGetProp(node, "name");
  xmlNode* cur;
    
  forEachChild(node, cur) {
    if( xmlNameEq(cur, "pointer") ) {
      // prefix pointer name with surface name
      char* p_name_orig = xmlGetProp(cur, "name");
      char* p_name = new char[strlen(name) + strlen(p_name_orig) + 2];
      sprintf(p_name, "%s_%s", name, p_name_orig);
      xmlSetProp(cur, "name", p_name);

      
      if( strcmp(p_name_orig, "a") == 0 )
        a = vrm->parsePointer(cur);
      else if( strcmp(p_name_orig, "b") == 0 )
        b = vrm->parsePointer(cur);
      else if( strcmp(p_name_orig, "c") == 0 )
        c = vrm->parsePointer(cur);
      else {
        msg(M_ERR, "Pointer named \"%s\" not allowed inside of <surface type=\"3point\">. Use one of a, b, c.\n", xmlGetProp(cur, "name"));
      }
    }
    else {
      msg(M_WARN, "Only <pointer> tags allowed inside <surface>. Ignoring \"%s\"\n", xmlNodeName(cur));
    }
  }
}


Surface_3Point::Surface_3Point(Pointer* a_new, Pointer* b_new, Pointer* c_new, VRManager* vrm, char* name) : Surface(name, vrm) {
  setPointers(a_new, b_new, c_new);
}

void Surface_3Point::setPointers(Pointer* a_new, Pointer* b_new, Pointer* c_new) {
  a = a_new;
  a->addDependent(this);
  b = b_new;
  b->addDependent(this);
  c = c_new;
  c->addDependent(this);
//  depChanged();
}

void Surface_3Point::depChanged(int msg) {
  bool changed = true;
}

void Surface_3Point::update() {
  VRManager* vrm = getVrm();
//  setPlane(getVrm()->getPointerLocation(a), getVrm()->getPointerLocation(b), getVrm()->getPointerLocation(c));
  setPlane(a->getLocation(), b->getLocation(), c->getLocation());
}





/*



mask* MaskSet::quadIntersect(quadPair p) {
  mask* result = new mask;
  
  
    int pts = 0;
    pt cur, last;
    cur.t = 0;
    cur.poly = 0;
    cur.vtx = 0;
    last = cur;
    pt first = cur;
    // find first point
    bool gotFirst = false;
// printf("######  Looking for first point..\n");      
    
    while( ! gotFirst ) {
      if( pointInOtherQuad(p, cur) >= 0 || cur.t != 0 ) {
// printf("######  Found first point!\n");      
        gotFirst = true;
        pts++;
        break;
      }
      else {
        cur = nextPt(p, cur, last);
        if( cur.poly == first.poly && cur.vtx == first.vtx && cur.t == first.t ) {
          if( cur.poly == 0 ) {
            cur.t = 0;
            cur.poly = 1;
            cur.vtx = 0;
            last = cur;
            first = cur;
          }
          else {
//             printf("No intersection!\n");
            return;
          }
        }
      }  
    }
    first = cur;
    
    bool done = false;
//     glColor4f(1, 1, 1, 0.5);
//     glBegin(GL_TRIANGLE_FAN);
//     doVertex(ptPos(cur));
    result->pts.push_front(ptPos(cur));
    
    
    
// printf("######  Starting search..\n");      
    while(! done) {
      pt next = nextPt(p, cur, last);
      last = cur;
      cur = next;        
//      if( cur.poly == first.poly && cur.vtx == first.vtx && cur.t == first.t ) {
      if( (ptPos(cur)-ptPos(first)).mag() < 0.01 ) {
        done = true;    
// printf("######  Found endpoint!\n");      
      }
      else {
        pts++;
        if( pts > 10 ) {
          printf("Way too many points! bailing out.\n");
          return NULL;
        }
//         doVertex(ptPos(cur));
        result->pts.push_front(ptPos(cur));
      }
    }
//     glEnd();

  
  return result;
}




  bool MaskSet::segmentIntersect(Vector a1, Vector a2, Vector b1, Vector b2, float* t1, float* t2, Vector* i) {
  
// printf("        ----- intersect:  ");
    float fpf = 0.00001;
    Vector ma = a2-a1;
    Vector mb = b2-b1;
    float tn = -a1[1]*mb[0] + b1[1]*mb[0] + (a1[0]-b1[0])*mb[1];
    float td = (ma[1]*mb[0]-ma[0]*mb[1]);
    float t = tn/td;
    if( i != NULL && td != 0 ) {
      i->set(a1 + (a2-a1) * t);
    }

    *t1 = t;
    if( mb[0] == 0 ) {
      *t2 = (a1[1] + (ma[1] * t) - b1[1]) / mb[1];
    }
    else {
      *t2 = (a1[0] + (ma[0] * t) - b1[0]) / mb[0];
    }
// printf("        t1 = %f   t2 = %f ", t, *t2);
    
    if( *t1 > 0-fpf && *t1 < 1+fpf && *t2 > 0-fpf && *t2 < 1+fpf ) {
// printf("  Yes.\n");
      return true;
    }
// printf("  No.\n");
    return false;
  }

  
    
  int MaskSet::pointWithinQuad(Vector ul, Vector ur, Vector bl, Vector br, Vector pt) {
    float fpf = 0.00001;
      // distort matrices are already calculated by the Viewports ..?
    Vector p = distortMatrix(ul, ur, bl, br).inverse() * pt;
    
    p /= p[3];
// printf("        ----- checking if point %f %f is in unit box.. (given %f %f)  ", p[0], p[1], pt[0], pt[1]);
    if( p[0] > 0-fpf && p[0] < 1+fpf && p[1] > 0-fpf && p[1] < 1+fpf ) {
// printf("Yes.\n");
      return 1;
    }
    else if( p[0] >= 0-fpf && p[0] <= 1+fpf && p[1] >= 0-fpf && p[1] <= 1+fpf ) {
// printf("On border.\n");
      return 0;
    }
    else {
// printf("No.\n");
      return -1;
    }
  }
  
  int MaskSet::pointWithinQuad(quadPair p, int q, Vector pt) {
    return pointWithinQuad(p[q][0], p[q][1], p[q][3], p[q][2], pt);     
  }
  

  pt MaskSet::nextPt(quadPair p, pt a, pt last) {
    pt next;
    
// printf("Finding next point for (%d, %d, %f), (%d, %d, %f)\n", a.poly, a.vtx, a.t, last.poly, last.vtx, last.t);
//    if( a.t == 0 || (last.t != 0 && a.it == 0) || (last.t == 0 && pointWithinQuad(rot(last.poly,2), ptPos(last)) < 0 && a.it != 0 ) ) {
    if( a.t == 0 || (last.t == 0 && pointWithinQuad(p, rot(last.poly,2), ptPos(last)) < 0 ) ) {
// printf("  appears to be at vertex or last point was outside.\n");
      // find nearest intersection
      float min_t = 1.0;
      next.poly = a.poly;
      next.t = 0;
      next.vtx = rot(a.vtx,4);
      pt nextVtx = next;
      pt prevVtx = a;
      prevVtx.t = 0;
      pt ipt, ipt2;
      ipt.poly = rot(a.poly, 2);
      ipt.t = 0;
      ipt2 = ipt;
      
// printf("  searching for intersecting point..\n");      
      for( int i=0; i<4; i++ ) {
        ipt.vtx = i;
        ipt2.vtx = rot(i, 4);
        float it, it2;
        bool intsct = segmentIntersect(ptPos(prevVtx), ptPos(nextVtx), ptPos(ipt), ptPos(ipt2), &it, &it2);
// printf("    checked %d %d - %d %d  X  %d %d - %d %d  = %f\n", prevVtx.poly, prevVtx.vtx, nextVtx.poly, nextVtx.vtx, ipt.poly, ipt.vtx, ipt2.poly, ipt2.vtx, it);
        if( intsct && it < min_t && it > a.t ) {
// printf("    found one at t=%f\n", it);      
          min_t = it;
          next.t = it;
          next.vtx = a.vtx;
          next.ivtx = ipt.vtx;
          next.it = it2;
        }
      }
      
      if( next.t == 1 ) {
        next.t = 0;
        next.vtx = rot(next.vtx, 4);
      }
      return next;
    }
    else {
// printf("  appears to be an intersection point..\n");      
      next.poly = rot(a.poly, 2);
      next.t = 0;
      next.vtx = a.ivtx;
      
      pt a_begin, a_end, b_end;
      a_begin = a_end = a;
      a_begin.t = 0;
      a_end.t = 0;
      a_end.vtx = rot(a.vtx, 4);
      b_end = next;
      b_end.vtx = rot(b_end.vtx,4);
      
      float t, t2;
      bool intsct = segmentIntersect(ptPos(next), ptPos(b_end), ptPos(a_begin), ptPos(a_end), &t, &t2);
// printf("    intersects at t=%f\n", t);      
      
      
      // find nearest points backward and forward
      
      pt bck, fwd;
      bck = fwd = next;
      bck.t = 0;
      fwd.t = 1;
// printf("    searching for points closer than endpoints..\n");      
      pt iStart, iEnd;
      iStart = iEnd = a;
      iStart.t = iEnd.t = 0;
      for( int i=0; i<3; i++ ) {
        iStart.vtx = (a.vtx + 1 + i) % 4;
        iEnd.vtx = (a.vtx + 2 + i) % 4;
        float it, it2;
        bool intsct = segmentIntersect(ptPos(next), ptPos(b_end), ptPos(iStart), ptPos(iEnd), &it, &it2);
// printf("    checked %d %d - %d %d  X  %d %d - %d %d  = %f\n", next.poly, next.vtx, b_end.poly, b_end.vtx, iStart.poly, iStart.vtx, iEnd.poly, iEnd.vtx, it);
        if( intsct && it > bck.t && it < t ) {
// printf("      ..found one for bck at t=%f\n", it);
          bck.t = it;
          bck.ivtx = iStart.vtx;
          bck.it = it2;          
        }
        if( intsct && it > t && it < fwd.t ) {
// printf("      ..found one for fwd at t=%f\n", it);
          fwd.t = it;
          fwd.ivtx = iStart.vtx;
          fwd.it = it2;
        }
      }
      
//float nt = pointInOtherQuad(next);
// printf("    t for bck = %f (%d) (%f, %f)\n", next.t, nt, (ptPos(next))[0], (ptPos(next))[1]);
      if( bck.t != 0 || pointInOtherQuad(p, bck) >= 0 ) {
// printf("    returning bck\n");      
        return next;
      }
      
      
      if( fwd.t == 1 ) {
        fwd.t = 0;
        fwd.vtx = rot(fwd.vtx, 4);
      }
//nt = pointInOtherQuad(next);
// printf("    t for fwd = %f (%d) (%f, %f)\n", next.t, nt, (ptPos(next))[0], (ptPos(next))[1]);
      if( fwd.t != 0 || pointInOtherQuad(p, fwd) >= 0 ) {
// printf("    returning fwd\n");      
        return fwd;
      }
      else {
        printf("Could not find next point!\n");
      }
    }
  }

  int MaskSet::pointInOtherQuad(quadPair p, int a, int b) {
    return pointWithinQuad(p, rot(a,2), p[a][b]); 
  }
  
  int MaskSet::pointInOtherQuad(quadPair p, pt a) {
    if( a.poly < 0 || a.poly > 1 || a.vtx < 0 || a.vtx > 3 ) {
      printf("Bad pt!\n");
    }
    return pointWithinQuad(p, rot(a.poly,2), ptPos(a));
  }

  
int MaskSet::rot(int a, int mod) {
  return (a+1) % mod;
}


Vector MaskSet::ptPos(pt a) {
  return p[a.poly][a.vtx] + (p[a.poly][rot(a.vtx,4)]-p[a.poly][a.vtx]) * a.t;
}*/



// MaskSet::MaskSet() {
//   invalidateMasks();
// }
// 
// void MaskSet::addViewport(Viewport* vp) {
//   char* name = vp->getVrmName();
//   maskGroup* tmp = maskHash[name];
//   if( tmp == NULL ) {
//     //create new
//     tmp = new maskHash;
//     maskHash[name] = tmp;
//     tmp->valid = false;
//     tmp->vp = vp;
//   }
//   else {
//     tmp->valid = false;
//   }
// }
// 
// void MaskSet::drawMask(char* name) {
//   if( ! maskHash[name]->valid ) {
//     updateMask(name);
//   }
//   
//   maskGroup* m = maskHash[name];
//   if( m == NULL ) {
//     return;
//   }
// 
//   
//   glDisable(GL_LIGHTING);
//   glDisable(GL_TEXTURE_2D);
//   glDisable(GL_TEXTURE_1D);
//   
//   for( map<const char*, mask*, ltstr>::iterator i = m->maskHash.begin(); i != m->maskHash.end(); i++ ) {
//     slist<maskPt> pts = (*i).second;
//     glBegin(GL_TRIANGLE_FAN);
//       for( slist<maskPt>::iterator j = pts.begin(); j != pts.end(); j++ ) {
//         glColor4f(0, 0, 0, (*j).a);
//         glVertex3f((*j).v[0], (*j).v[1], (*j).v[2]);      
//       }
//     glEnd();
//   }
// 
// }
// 
// void MaskSet::depChanged(int msg=0) {
//   invalidateMasks();
// }
// 
// void MaskSet::invalidateMasks() {
//   for( map<const char*, maskGroup*, ltstr>::iterator i = maskHash.begin(); i != maskHash.end(); i++ ) {
//     (*i)->valid = false;
//   }
// }
// 
// void MaskSet::updateMask(char* name) {
//   maskGroup* m = maskHash[name];
//   if( m == NULL ) {
//     return;
//   }
//   
//   for( map<const char*, maskGroup*, ltstr>::iterator i = maskHash.begin(); i!= maskHash.end(); i++ ) {
//     if( *i != m ) {
//     
//       quadPair tmp;
//       tmp[0][0] = m->vp->getCorner("ul")->getLocation();
//       tmp[0][1] = m->vp->getCorner("ur")->getLocation();
//       tmp[0][2] = m->vp->getCorner("bl")->getLocation();
//       tmp[0][3] = m->vp->getCorner("br")->getLocation();
//       
//       tmp[1][0] = (*i)->vp->getCorner("ul")->getLocation();
//       tmp[1][1] = (*i)->vp->getCorner("ur")->getLocation();
//       tmp[1][2] = (*i)->vp->getCorner("bl")->getLocation();
//       tmp[1][3] = (*i)->vp->getCorner("br")->getLocation();
//     
//       mask* mptr = new mask;
//       mptr = m->maskHash[(*i)->vp->getVrmName()];
//       delete mptr;
//       
//         // Check for existing inverse 
//       *mptr = quadIntersect(tmp)      
//       m->maskHash[(*i)->vp->getVrmName()] = mptr;
//     }
//   }
// }
