// Copyright 2004,2005 Luke Campagnola <lcampagn@mines.edu>, Colorado School of Mines
// 
// This file is part of VRManager.
// 
// VRManager is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
// 
// VRManager is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with VRManager; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



#include "TrackerSystem.h"
#include "debug.h"
#include "vrmanager.h"
  // TrackerSystem methods


TrackerSystem* createNewTrackerSystem(xmlNode* node, VRManager* vrm) {
  TrackerSystem* ret;
  
  Matrix cs;
  assertNodeHasProp(node, "type");
  char* type = xmlGetProp(node, "type");

  if( xmlPropEq(node, "type", "vrpn_tracker") ) {
    #ifdef HAVE_VRPN
      ret = new TS_VrpnTracker(node, vrm);
    #else
      msg(M_WARN, "VRPN not supported. Recompile with -DHAVE_VRPN\n");
    #endif
  }
  else if( xmlPropEq(node, "type", "vrpn_button") ) {
    #ifdef HAVE_VRPN
      ret = new TS_VrpnButton(node, vrm);
    #else
      msg(M_WARN, "VRPN not supported. Recompile with -DHAVE_VRPN\n");
    #endif
  }
  else {
    msg(M_WARN, "Tracker system type \"%s\" not supported!\n", xmlGetProp(node, "type"));
  }
    
  ret->setVrmName(xmlGetProp(node, "name"));
  ret->setVrm(vrm);
  return ret;
}



  
    
TrackerSystem::TrackerSystem() {
}  

TrackerSystem::~TrackerSystem() {
}  

// void TrackerSystem::addTracker(const char* name, Tracker* tracker) {
//   tracker->setCoordSystem(coordSystem);
//   trackerHash[name] = tracker;
// }

Tracker* TrackerSystem::getTracker(const char* name) {
  return trackerHash[name];
}



void TrackerSystem::setButtonCallback(void (*callbackFn)(int, int, void*), void* userData) {
  buttonCallback = callbackFn;
  buttonCbUserData = userData;
}

  
void TrackerSystem::setCoordSystem(Matrix newCS) {
  coordSystem = newCS;
}

Tracker* TrackerSystem::parseTracker(xmlNode* node) {
  msg(M_INFO, "  == Parsing tracker %s\n", xmlGetProp(node, "name"));

  assertNodeHasProp(node, "name");
  char *name = xmlGetProp(node, "name");

  Tracker* tkr = getTracker(name);
  if( tkr == NULL ) {
    tkr = new Tracker();
  }
  else {
    msg(M_WARN, "Duplicate tracker name \"%s\"; ignoring.\n", name);
    return NULL;
  }
  
/*    // Check for elements internal to the tracker such as tracked pointers
  xmlNode *node1 = NULL;
  forEachChild(node, node1) {

      // If there is a pointer linked to the tracker, make sure it is given the correct parameters
    if( xmlNameEq(node1, "pointer") ) {
      if( ! nodeHasProp(node1, "type") ) {
        xmlSetProp(node1, "type", "tracked");
      }
      if( xmlPropEq(node1, "type", "tracked")) {
        if( ! nodeHasProp(node1, "trackerSystem") ) {
          xmlSetProp(node1, "trackerSystem", TSName);
        }
        if( ! nodeHasProp(node1, "tracker") ) {
          xmlSetProp(node1, "tracker", name);
        }
      }
      vrm->parsePointer(node1);
    }
  }*/
  return tkr;
}





  // VRPN Tracker Types

#ifdef HAVE_VRPN



TS_VrpnTracker::TS_VrpnTracker() {
}

TS_VrpnTracker::TS_VrpnTracker(xmlNode* node, VRManager* vrm) {
  reconfigure(node, vrm);
}

TS_VrpnButton::TS_VrpnButton() {
}

TS_VrpnButton::TS_VrpnButton(xmlNode* node, VRManager* vrm) {
  reconfigure(node, vrm);
}


void TS_VrpnTracker::reconfigure(xmlNode* node, VRManager* vrm) {
  // TODO: Handle old threads!
  assertNodeHasProp(node, "vrpn_name");
  

  type = TSTYPE_VRPN_TRACKER;
  buttonCallback = NULL;
  trackerName = xmlGetProp(node, "vrpn_name");
    
  // Open the tracker
  vrpnTkr = new vrpn_Tracker_Remote(trackerName);

  // Set up the tracker callback handler
  vrpnTkr->register_change_handler((void*) this, handleVrpnTrackerWrapper);

  // Start thread to poll for VRPN events
  gotVrpnReport = false;
  vrpnPollInterval = 10000;  // Check for vrpn updates every 10ms
  pthread_create( &vrpnThread, NULL, vrpnTrackerUpdateWrapper, (void*) this);

  
  xmlNode *cur = NULL;
  forEachChild(node, cur) {
    if( xmlNameEq(cur, "coordSystem") ) {
      msg(M_DEBUG, "  Adding coordinate system.\n");
      setCoordSystem(vrm->xmlCsMatrix(cur));
    }
    else if( xmlNameEq(cur, "tracker") ) {
/*      xmlSetProp(cur, "trackerSystem", trackerName);*/
//      xmlSetProp(cur, "trackerSystemType", xmlGetProp(node, "type"));
      assertNodeHasProp(cur, "vrpn_id");
      Tracker* t = parseTracker(cur);
      addTracker(t, xmlGetProp(cur, "name"), xmlGetProp(cur, "vrpn_id"));
      
      // Pass on information to children, then move past the end of the trackerSystem
      if( nodeHasChildren(cur) ) {
        setChildrenProp(cur, NULL, "trackerSystem", xmlGetProp(node, "name"));
        setChildrenProp(cur, NULL, "tracker", xmlGetProp(cur, "name"));
        moveChildrenAfterNode(cur, node);
      }
    }
  }
}

void TS_VrpnButton::reconfigure(xmlNode* node, VRManager* vrm) {
  // TODO: Handle old threads!
  
  
  assertNodeHasProp(node, "vrpn_name");

  type = TSTYPE_VRPN_BUTTON;
  buttonCallback = NULL;
  trackerName = xmlGetProp(node, "vrpn_name");
  
// Open the tracker
  vrpnBtn = new vrpn_Button_Remote(trackerName);

// Set up the tracker callback handler
  vrpnBtn->register_change_handler((void*) this, handleVrpnButtonWrapper);
  
// Start thread to poll for VRPN events
  gotVrpnReport = false;
  vrpnPollInterval = 10000;  // Check for vrpn updates every 10ms
  pthread_create( &vrpnThread, NULL, vrpnButtonUpdateWrapper, (void*) this);
}

void TS_VrpnTracker::cleanup() {
  threadExit = true;
  msg(M_DEBUG, "Waiting for tracker poll thread to exit..\n");
  pthread_join(vrpnThread, NULL);
}

void TS_VrpnButton::cleanup() {
  threadExit = true;
  msg(M_DEBUG, "Waiting for tracker poll thread to exit..\n");
  pthread_join(vrpnThread, NULL);
}

void TS_VrpnTracker::addTracker(Tracker* tracker, const char* name, char* vrpn_id) {
  tracker->setCoordSystem(coordSystem);
  char* tmp = new char[strlen(name)+1];
  strcpy(tmp, name);
  trackerHash[tmp] = tracker;
  trackerIdHash[vrpn_id] = tracker;
}


void TS_VrpnTracker::handleVrpnTrackerWrapper(void* ts, const vrpn_TRACKERCB t) {
  TS_VrpnTracker* tPtr = (TS_VrpnTracker*) ts;
  tPtr->handleVrpnTracker(t);
}

void TS_VrpnTracker::handleVrpnTracker(const vrpn_TRACKERCB t) {
  // t.sensor, t.pos, t.quat
  char* id = new char[100];
  sprintf(id, "%d", (int)t.sensor);
  Tracker *tkr = trackerIdHash[id];
    // Update tracker
  tkr->setLocation(Vector(t.pos[0], t.pos[1], t.pos[2]));
  
  q_matrix_type m;
  q_to_col_matrix(m, t.quat);
  Matrix rotation(m);
  tkr->setOrientationMatrix(rotation);
  
  gotVrpnReport = true;
}


void TS_VrpnButton::handleVrpnButtonWrapper(void* ts, vrpn_BUTTONCB b) {
  TS_VrpnButton* tPtr = (TS_VrpnButton*) ts;
  tPtr->handleVrpnButton(b);
}

  
void TS_VrpnButton::handleVrpnButton(vrpn_BUTTONCB b) {
  if( buttonCallback != NULL ) {
    buttonCallback(b.button, b.state, buttonCbUserData);
  }
}


void *TS_VrpnTracker::vrpnTrackerUpdateWrapper(void* ts) {
  TS_VrpnTracker* tPtr = (TS_VrpnTracker*) ts;
  tPtr->vrpnTrackerUpdate();
}

void TS_VrpnTracker::vrpnTrackerUpdate() {
  while( ! threadExit ) {
    vrpnTkr->mainloop();  // This is just here to flush out the event queue
  
      // wait for a new report
    gotVrpnReport = false;
    int c = 0;
    while( ! gotVrpnReport && c < 100 ) {
      vrpnTkr->mainloop();
      c++;
      usleep(5000);   // wait 5ms before checking again
    }
    
    usleep(vrpnPollInterval);
  }
}

void *TS_VrpnButton::vrpnButtonUpdateWrapper(void* ts) {
  TS_VrpnButton* tPtr = (TS_VrpnButton*) ts;
  tPtr->vrpnButtonUpdate();
}

void TS_VrpnButton::vrpnButtonUpdate() {
  while( ! threadExit ) {
    vrpnBtn->mainloop();
    usleep(vrpnPollInterval);
  }
}

#endif
