// Copyright 2004,2005 Luke Campagnola <lcampagn@mines.edu>, Colorado School of Mines
// 
// This file is part of VRManager.
// 
// VRManager is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
// 
// VRManager is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with VRManager; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include <stdio.h>
#include <GL/gl.h>
#include "Viewport.h"
#include "debug.h"
#include "vrmanager.h"
#include "xmlHelpers.h"
  
Viewport* createNewViewport(xmlNode* node, class VRManager* vrm, char* name) {
  Viewport* ret;
  if( ! nodeHasProp(node, "type") ) {
    if( nodeHasProp(node, "alias") ) {
      xmlSetProp(node, "type", "alias");
    }
    else {
      xmlSetProp(node, "type", "static");
    }
  }
  
  
  assertNodeHasProp(node, "type");
  if( xmlPropEq(node, "type", "static") || xmlPropEq(node, "type", "projector") ) {
    ret = new Viewport(node, vrm);
  }
  else if( xmlPropEq(node, "type", "alias") ) {
    ret = new Viewport_Alias(node, vrm);
  }
  else {
    msg(M_WARN, "Viewport type \"%s\" not supported.\n", xmlGetProp(node, "type"));
  }
  
    // Set flags for distortion handling
  bool forceRectangular = false;
  bool fastDistortion = false;
  if( nodeHasProp(node, "forceRectangular") && xmlPropEq(node, "forceRectangular", "true") ) {
    forceRectangular = true;
  }
  if( nodeHasProp(node, "fastDistortion") && xmlPropEq(node, "fastDistortion", "true") ) {
    fastDistortion = true;
  }
  ret->setForceRectangular(forceRectangular);
  ret->setFastDistortion(fastDistortion);
  
    // Set polarization
  if( nodeHasProp(node, "polarization") && xmlPropEq(node, "polarization", "right") ) {
    ret->setPolarization(1);
  }
  if( nodeHasProp(node, "drawArea") ) {
    char* str = new char[strlen(xmlGetProp(node, "drawArea"))+1];
    strncpy(str, xmlGetProp(node, "drawArea"), strlen(xmlGetProp(node, "drawArea")));
    char* tok;
    tok = strtok(str, ",");
    for( int i=0; i<4; i++ ) {
      if( tok == NULL ) {
        msg(M_ERR, "viewport drawArea attribute must have 4 comma-separated integers \"top, left, height, width\"\n");
      }
      ret->drawArea[i] = atoi(tok);
      tok = strtok(NULL, ",");
    }

  }
  
    // Set surface (If no surface is explicitly specified, a new one will be set later)
  if( nodeHasProp(node, "surface") ) {
    Surface* s = vrm->getSurface(xmlGetProp(node, "surface"));
    ret->setSurface(s);
    s->addViewport(ret);
  }
  
  
  ret->setVrmName(name);
  ret->setVrm(vrm);
  return ret;
}


  
  // Viewport methods

void Viewport::init() {
  drawArea[0] = drawArea[1] = drawArea[2] = drawArea[3] = -1;
  pm_valid = false;
  cs_valid = false;
  forceRectangular = false;
  useDistortion = true;
  allowBlending = true;
  surface = NULL;
}
  
Viewport:: Viewport() {
  init();
}

Viewport::Viewport(Pointer* uln, Pointer* urn, Pointer* bln, Pointer* brn, int p) {
  init();
  polarization = p;
  setCorners(uln, urn, bln, brn);
}

Viewport::Viewport(xmlNode* node, class VRManager* vrm) {
  init();
  reconfigure(node, vrm);
}  

int* Viewport::getDrawArea() {
  return drawArea;
}

void Viewport::reconfigure(xmlNode* node, class VRManager* vrm) {
  char* name = xmlGetProp(node, "name");
  
  
    // If the viewport is tracked, pass the info on to its pointers
  bool tracked = false;
  char* trackerSystem = NULL;
  char* tracker = NULL;

  if( nodeHasProp(node, "trackerSystem") ) {
    tracked = true;
    trackerSystem = xmlGetProp(node, "trackerSystem");
    if( trackerSystem == NULL ) {
      msg(M_ERR, "trackerSystem \"%s\" not found\n", xmlGetProp(node, "trackerSystem"));
      tracked = false;
    }
    tracker = xmlGetProp(node, "tracker");
    if( tracker == NULL ) {
      msg(M_ERR, "tracker \"%s\" not found in trackerSystem \"%s\"\n", xmlGetProp(node, "tracker"), xmlGetProp(node, "trackerSystem"));
      tracked = false;
    }
  }
  

  Pointer *ul, *ur, *bl, *br;
  ul = ur = bl = br = NULL;
  
    // Gather information about pointers to corners
  if( xmlPropEq(node, "type", "static") ) {
    if( nodeHasChildren(node) ) {
      // Parse all children here, make sure we get one for each corner.
      // Also, prefix each pointer's name with the viewportName_
      xmlNode *cur = NULL;
      forEachChild(node, cur) {
        if( xmlNameEq(cur, "pointer") ) {
          char* p_name_orig = xmlGetProp(cur, "name");
          char* p_name = new char[strlen(name) + 4];
          sprintf(p_name, "%s_%s", name, p_name_orig);
          xmlSetProp(cur, "name", p_name);

            // apply tracker info
          if( tracked ) {
            xmlSetProp(cur, "type", "tracked");
            xmlSetProp(cur, "trackerSystem", trackerSystem);
            xmlSetProp(cur, "tracker", tracker);
          }


          if( strcmp(p_name_orig, "ul") == 0 )
            ul = vrm->parsePointer(cur);
          else if( strcmp(p_name_orig, "ur") == 0 )
            ur = vrm->parsePointer(cur);
          else if( strcmp(p_name_orig, "bl") == 0 )
            bl = vrm->parsePointer(cur);
          else if( strcmp(p_name_orig, "br") == 0 )
            br = vrm->parsePointer(cur);
          else {
            msg(M_ERR, "Pointer named \"%s\" not allowed inside of <viewport>. Use one of ul, ur, bl, br.\n", xmlGetProp(cur, "name"));
          }
        }
        else {
          msg(M_WARN, "Only <pointer> tags allowed inside <viewport>. Ignoring \"%s\"\n", xmlNodeName(cur));
        }
      }
    }
    else {
      // Pointers are defined by attributes
      assertNodeHasProp(node, "ul");
      assertNodeHasProp(node, "ur");
      assertNodeHasProp(node, "bl");
      assertNodeHasProp(node, "br");
      ul = vrm->getPointer(xmlGetProp(node, "ul"));
      ur = vrm->getPointer(xmlGetProp(node, "ur"));
      bl = vrm->getPointer(xmlGetProp(node, "bl"));
      br = vrm->getPointer(xmlGetProp(node, "br"));
    }
  
    if( ul == NULL || ur == NULL || bl == NULL || br == NULL ) {
      msg(M_ERR, "Missing 1 or more corners for viewport \"%s\"\n", name);
    }
  }
  else if( xmlPropEq(node, "type", "projector") ) {
    Pointer *uln, *urn, *bln, *brn;
    Pointer *ulf, *urf, *blf, *brf;
    uln = urn = bln = brn = NULL;
    ulf = urf = blf = brf = NULL;


      // get the surface
    assertNodeHasProp(node, "surface");
    Surface* surface = vrm->getSurface(xmlGetProp(node, "surface"));
    if( surface == NULL ) {
      msg(M_ERR, "No surface named \"%s\"\n", xmlGetProp(node, "surface"));
    }

      // First read in the eight frustum points

    if( nodeHasChildren(node) ) {
      // Parse all children here, make sure we get one for each corner.
      // Also, prefix each pointer's name with the viewportName_
      xmlNode *cur = NULL;
      forEachChild(node, cur) {
        if( xmlNameEq(cur, "pointer") ) {
          char* p_name_orig = xmlGetProp(cur, "name");
          char* p_name = new char[strlen(name) + strlen(p_name_orig) + 2];
          sprintf(p_name, "%s_%s", name, p_name_orig);
          xmlSetProp(cur, "name", p_name);

            // apply tracker info
          if( tracked ) {
            xmlSetProp(cur, "type", "tracked");
            xmlSetProp(cur, "trackerSystem", trackerSystem);
            xmlSetProp(cur, "tracker", tracker);
          }

          if( strcmp(p_name_orig, "uln") == 0 )
            uln = vrm->parsePointer(cur);
          else if( strcmp(p_name_orig, "urn") == 0 )
            urn = vrm->parsePointer(cur);
          else if( strcmp(p_name_orig, "bln") == 0 )
            bln = vrm->parsePointer(cur);
          else if( strcmp(p_name_orig, "brn") == 0 )
            brn = vrm->parsePointer(cur);
          else if( strcmp(p_name_orig, "ulf") == 0 )
            ulf = vrm->parsePointer(cur);
          else if( strcmp(p_name_orig, "urf") == 0 )
            urf = vrm->parsePointer(cur);
          else if( strcmp(p_name_orig, "blf") == 0 )
            blf = vrm->parsePointer(cur);
          else if( strcmp(p_name_orig, "brf") == 0 )
            brf = vrm->parsePointer(cur);
          else {
            msg(M_ERR, "Pointer named \"%s\" not allowed inside of <viewport>. Use one of ul, ur, bl, br.\n", xmlGetProp(cur, "name"));
          }
        }
        else {
          msg(M_WARN, "Only <pointer> tags allowed inside <viewport>. Ignoring \"%s\"\n", xmlNodeName(cur));
        }
      }
    }
    else {
      // Pointers are defined by attributes
      assertNodeHasProp(node, "uln");
      assertNodeHasProp(node, "urn");
      assertNodeHasProp(node, "bln");
      assertNodeHasProp(node, "brn");
      assertNodeHasProp(node, "ulf");
      assertNodeHasProp(node, "urf");
      assertNodeHasProp(node, "blf");
      assertNodeHasProp(node, "brf");
      uln = vrm->getPointer(xmlGetProp(node, "uln"));
      urn = vrm->getPointer(xmlGetProp(node, "urn"));
      bln = vrm->getPointer(xmlGetProp(node, "bln"));
      brn = vrm->getPointer(xmlGetProp(node, "brn"));
      ulf = vrm->getPointer(xmlGetProp(node, "ulf"));
      urf = vrm->getPointer(xmlGetProp(node, "urf"));
      blf = vrm->getPointer(xmlGetProp(node, "blf"));
      brf = vrm->getPointer(xmlGetProp(node, "brf"));
    }

    if( uln == NULL || urn == NULL || bln == NULL || brn == NULL || ulf == NULL || urf == NULL || blf == NULL || brf == NULL ) {
      msg(M_ERR, "Missing 1 or more corners for projector frustum \"%s\"\n", name);
    }

      // create intersection points with the named Surface
    ul = new Pointer_Intersection(uln, ulf, surface);
    ur = new Pointer_Intersection(urn, urf, surface);
    bl = new Pointer_Intersection(bln, blf, surface);
    br = new Pointer_Intersection(brn, brf, surface);
  }
  setCorners(ul, ur, bl, br);
}


void Viewport::depChanged(int msg) {
  if( msg == 0 ) {
    pm_valid = false;
  }
  else {
    cs_valid = false;
    pm_valid = false;
  }
}

void Viewport::setCorners(Pointer* uln, Pointer* urn, Pointer* bln, Pointer* brn) {
  ul_pointer = uln;  
  ur_pointer = urn;  
  bl_pointer = bln;  
  br_pointer = brn;
  uln->addDependent(this, 1);
  urn->addDependent(this, 1);
  bln->addDependent(this, 1);
  brn->addDependent(this, 1);
}

Vector* Viewport::getCenter() {
  return &center;
}
  
Vector* Viewport::getNormal() {
  return &normal;
}


void Viewport::update() {

    // Copy the screen corner locations to local variables
  ur = ur_pointer->getLocation();
  ul = ul_pointer->getLocation();
  br = br_pointer->getLocation();
  bl = bl_pointer->getLocation();
  
    // Find coordinate system for screen
  normal = ((bl-ul) * (ur-ul)) + ((ur-br) * (bl-br)).norm();
  Vector up = ((ul-bl) + (ur-br)).norm();
  Vector rt = (up * normal).norm();
  center = (ul + ur + bl + br) * 0.25;
  
    // Find corners in local coordinate system -- screen is in xy plane, centered at origin
  Matrix cs = coordChangeMatrix(Vector(), Vector(1, 0, 0), Vector(0, 1, 0), center, rt, up);
  Vector ul_l, ur_l, bl_l, br_l;  
  ul_l = cs * ul;
  ur_l = cs * ur;
  bl_l = cs * bl;
  br_l = cs * br;
  
    // Find bounding box      
  float l, r, b, t;
  l = r = t = b = 0;
  l = ul_l[0] < bl_l[0] ? ul_l[0] : bl_l[0];
  r = ur_l[0] > br_l[0] ? ur_l[0] : br_l[0];
  b = bl_l[1] < br_l[1] ? bl_l[1] : br_l[1];
  t = ul_l[1] > ur_l[1] ? ul_l[1] : ur_l[1];
  
  
  // Create distortion matrix
    // find matrix that scales bounding box to unit box
  Matrix ubm = scaleMatrix(Vector(1.0 / (r-l), 1.0 / (t-b), 1)) * translateMatrix(Vector(-l, -b, 0));
    // transform corners to fit within unit box
  ul_u = ubm * ul_l;
  ur_u = ubm * ur_l;
  bl_u = ubm * bl_l;
  br_u = ubm * br_l;
  
  if( forceRectangular ) {
    distortionMatrix = Matrix();
  }
  else {
    Matrix dm = distortMatrix(ul_u, ur_u, bl_u, br_u).inverse();
//    Matrix dm = realDistortMatrix(ul_u, ur_u, bl_u, br_u);
    distortionMatrix = translateMatrix(Vector(-1, -1, 0)) * scaleMatrix(Vector(2, 2, 1)) * dm * scaleMatrix(Vector(0.5, 0.5, 1)) * translateMatrix(Vector(1, 1, 0));
  }
  
  
    // Set some values that can be used to make frustum later
  coordSystem = cs;
  bb_left = l;
  bb_right = r;
  bb_top = t;
  bb_bottom = b;
  cs_valid = true;
}


Matrix Viewport::generateProjectionMatrix(Pointer* eye, float near, float far) {
    // THIS SHOULD NOT BE HERE
  if( ! cs_valid ) {
    update();
  }


    // First move the eye to the screen's coordinate system
  Vector eye_screen = coordSystem * eye->getLocation();
  
    // Generate some frustum parameters
  float n = near / fabs(eye_screen[2]);
  float l = n * (bb_left   - eye_screen[0]);
  float r = n * (bb_right  - eye_screen[0]);
  float b = n * (bb_top    - eye_screen[1]);
  float t = n * (bb_bottom - eye_screen[1]);
    
    // Create the frustum Matrix: frustum * point down -Z axis * move camera to origin
    // THIS CAN EASILY BE OPTIMIZED MORE
  Matrix frustMatrix = frustumMatrix(l, r, t, b, near, far) * translateMatrix(Vector(-eye_screen[0], -eye_screen[1], -eye_screen[2]));
  
  
    // Rotate/translate the frustum to sit on top of the screen
  if( useDistortion ) {
    return distortionMatrix * frustMatrix * coordSystem;
  }
  else {
    return frustMatrix * coordSystem;
  }
}

Matrix Viewport::generateOrthoMatrix(float near, float far) {
    // THIS SHOULD NOT BE HERE
  if( ! cs_valid ) {
    update();
  }
  
  Matrix orthMatrix = orthoMatrix(bb_left, bb_right, bb_bottom, bb_top, near, far);
  
  
    // Rotate/translate the frustum to sit on top of the screen
  if( useDistortion ) {
//printf("using fast distort\n");
    return distortionMatrix * orthMatrix * coordSystem;
  }
  else {
//printf("using 2pass distort\n");
    return orthMatrix * coordSystem;
  }
}


void Viewport::correctDistortion(int x, int y, int w, int h) {
    // TODO: use glGet to determine active viewing area

  int tex_w = 2048;
  int tex_h = 2048;

    // Only distort if the projection matrix isn't doing it for us
    // and we don't have a rectangular screen
  if( ! useDistortion && ! forceRectangular ) {
      // Copy framebuffer to texture
    GLuint tex;
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
    glEnable(GL_TEXTURE_2D);
    glGenTextures(1, &tex);
    glBindTexture(GL_TEXTURE_2D, tex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, tex_w, tex_h, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
    glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, x, y, w, h, 0);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
      // TODO: use glPushPropib here
    
      
      // Render corrected shape
    glDisable(GL_LIGHTING);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_ALPHA);
    glDisable(GL_BLEND);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
      glLoadIdentity();
      glMatrixMode(GL_PROJECTION);
      glPushMatrix();
        glLoadIdentity();
        glOrtho(-1, 1, -1, 1, -1, 1);
        glDisable(GL_TEXTURE_2D);
          // Clear out previous buffer
        glColor4f(0, 0, 0, 1);
        glBegin(GL_QUADS);
          glVertex2f(-1, -1);
          glVertex2f(-1, 1);
          glVertex2f(1, 1);
          glVertex2f(1, -1);
        glEnd();
          // Re-render with perspective correction
        glLoadIdentity();
        glOrtho(-1, 1, -1, 1, -1, 1);
        glEnable(GL_TEXTURE_2D);
        glMultMatrixf(distortionMatrix.glMatrix());
        glBegin(GL_QUADS);
          glTexCoord2f(0, 0);
          glVertex2f(-1, -1);
          glTexCoord2f(0, 1);
          glVertex2f(-1, 1);
          glTexCoord2f(1, 1);
          glVertex2f(1, 1);
          glTexCoord2f(1, 0);
          glVertex2f(1, -1);
        glEnd();
      glPopMatrix();
      glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
    
    // Delete texture
    glDeleteTextures(1, &tex);
  }
}

  // draw the blending mask
void Viewport::drawMask() {
  if( surface != NULL ) {
    surface->drawMask(this);
  }
}


Pointer* Viewport::getCorner(int c) {
  if( c == 0 )
    return ul_pointer;
  else if( c == 1 )
    return ur_pointer;
  else if( c == 2 )
    return bl_pointer;
  else if( c == 3 )
    return br_pointer;
  else {
    msg(M_WARN, "Viewport has no corner number %d (must be 0-3)\n", c);
    return NULL;
  }
}


Pointer* Viewport::getCorner(char* c) {
  if( strcmp(c, "ul") == 0 )
    return getCorner(0);
  else if( strcmp(c, "ur") == 0 )
    return getCorner(1);
  else if( strcmp(c, "bl") == 0 )
    return getCorner(2);
  else if( strcmp(c, "br") == 0 )
    return getCorner(3);
  else {
    msg(M_WARN, "Viewport has no corner called \"%s\"\n", c);
    return NULL;
  }
}

void Viewport::setFastDistortion(bool b) {
  useDistortion = b;
}

void Viewport::setForceRectangular(bool b) {
  forceRectangular = b;  
}

void Viewport::setPolarization(int p) {
  polarization = p;
}

void Viewport::setSurface(Surface* s) {
  surface = s;
}

Surface* Viewport::getSurface() {
  return surface;
}

int Viewport::getPolarization() {
  return polarization;
}



//// Alias viewport

Viewport_Alias::Viewport_Alias() {
}

Viewport_Alias::Viewport_Alias(xmlNode* node, class VRManager* vrm) {
  reconfigure(node, vrm);
}

void Viewport_Alias::reconfigure(xmlNode* node, class VRManager* vrm) {
  assertNodeHasProp(node, "alias");
  alias = vrm->getViewport(xmlGetProp(node, "alias"));
  if( alias == NULL ) {
    msg(M_ERR, "Cannot make alias of non-existing viewport \"%s\"\n", xmlGetProp(node, "alias"));
  }
  else {
    alias->addDependent(this);
  }
}

Matrix Viewport_Alias::generateProjectionMatrix(Pointer* eye, float near, float far) {
  return alias->generateProjectionMatrix(eye, near, far);
}

Matrix Viewport_Alias::generateOrthoMatrix(float near, float far) {
  return alias->generateOrthoMatrix(near, far);
}

void Viewport_Alias::setCorners(Pointer* uln, Pointer* urn, Pointer* bln, Pointer* brn) {
  alias->setCorners(uln, urn, bln, brn);
}

void Viewport_Alias::correctDistortion(int x, int y, int w, int h) {
  alias->correctDistortion(x, y, w, h);
}

void Viewport_Alias::drawMask() {
  alias->drawMask();
}

Pointer* Viewport_Alias::getCorner(int c) {
  return alias->getCorner(c);
}

Pointer* Viewport_Alias::getCorner(char* c) {
  return alias->getCorner(c);
}


void Viewport_Alias::setFastDistortion(bool s) {
  alias->setFastDistortion(s);
}

void Viewport_Alias::setForceRectangular(bool s) {
  alias->setForceRectangular(s);
}

void Viewport_Alias::setPolarization(int p) {
  alias->setPolarization(p);
}

void Viewport_Alias::setSurface(Surface* s) {
  alias->setSurface(s);
}

Surface* Viewport_Alias::getSurface() {
  return alias->getSurface();
}

Vector* Viewport_Alias::getCenter() {
  return alias->getCenter();
}

Vector* Viewport_Alias::getNormal() {
  return alias->getNormal();
}

int Viewport_Alias::getPolarization() {
  return alias->getPolarization();
}

int* Viewport_Alias::getDrawArea() {
  return alias->getDrawArea();
}

