// Copyright 2003,2005 Luke Campagnola <lcampagn@mines.edu>, Colorado School of Mines
// 
// This file is part of libvectormath.
// 
// libvectormath is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
// 
// libvectormath is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with libvectormath; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


/***************************************
/ Vector and Matrix classes
/ Luke Campagnola (lcampagn@mines.edu)
/ 2003.10.15
***************************************/

#ifndef VECTORMATH_H
#define VECTORMATH_H


#include <math.h>
//#include <iostream.h>

#ifndef PI
#define PI 3.14159265358979323846
#endif

// degree/radian conversions
#ifndef R2D
#define R2D 57.2958
#endif 
#ifndef D2R
#define D2R 0.0174533
#endif



class Vector
{
  private:
  
    float x[4];
  
  public: 

 
  Vector(float=0, float=0, float=0, float=1);
  ~Vector();

  float& operator[] ( int index );
  float operator[] ( int index ) const;
  void set(float=0, float=0, float=0, float=1);
  void set( const Vector& op );

  // Vector-Vector operations
    
  Vector operator+ ( const Vector& op ) const;
  Vector operator- ( const Vector& op ) const;
  Vector operator- () const;
  void operator+= ( const Vector& op );
  void operator-= ( const Vector& op );
  Vector operator* ( const Vector& op ) const;  // Cross product
  void operator*= ( const Vector& op );
  float operator& ( const Vector& op ) const;  // Dot product
  float dot( const Vector& op ) const;  
  Vector proj( const Vector& op ) const;  
  float angle( const Vector& op ) const;  

  // Vector-scalar operations

  Vector operator* ( float scl ) const;
  Vector operator/ ( float scl ) const;
  void operator*= ( float scl );
  void operator/= ( float scl );

  // Vector operations

  Vector norm() const;  
  void normalize();  
  float mag() const;
  void print() const;

};

Vector lineLineMidpoint(Vector s1, Vector e1, Vector s2, Vector e2);


/*
ostream operator<< ( ostream out, Vector a ) {
  out << a.x[0] << " " << a.x[1] << " " << a.x[2] << " " << a.x[3];
  return out;
}
*/


class Matrix {
  private:

    float glMat[16];
    float x[4][4];  // row, column

  
  public:


    Matrix();
    Matrix(const double m[4][4]);
    Matrix(const float m[4][4]);

    float& operator() ( int r, int c );
    float operator() ( int r, int c ) const;
    
    Matrix operator+ (const Matrix& a) const;
    Matrix operator* (const Matrix& a) const;
    Matrix operator* (float a) const;
    Matrix operator/ (float a) const;
    Vector operator* (const Vector& a) const;
    float* glMatrix();
    Matrix inverse() const;
    Matrix tr() const;
    float det() const;    
    float subDet(int i, int j) const;
            
    void print() const;

};



Matrix translateMatrix(Vector a);
Matrix scaleMatrix(Vector a);
Matrix scaleMatrix(float a);
Matrix rotateMatrix(int axis, float angle);
Matrix rotateMatrix(Vector axis, float angle);
Matrix frustumMatrix(float l, float r, 
                     float b, float t, 
                     float n, float f);                     
Matrix orthoMatrix(float l, float r, 
                     float b, float t, 
                     float n, float f);                     
Matrix coordChangeMatrix(Vector o1, Vector x1, Vector y1, 
                         Vector o2, Vector x2, Vector y2); 
                         
                         
  // Returns a distortion matrix that morphs the unit box into the quad specified.
  // Note: this matrix works for opengl on its own, but if you want to use it 
  // with vectors, you'll need to say
  //    v /= v[3];
  // in order to get the correct 3d values
Matrix distortMatrix(Vector ul, Vector ur, Vector bl, Vector br);
Matrix realDistortMatrix(Vector ul, Vector ur, Vector bl, Vector br);

#endif
