// Copyright 2004,2005 Luke Campagnola <lcampagn@mines.edu>, Colorado School of Mines
// 
// This file is part of VRManager.
// 
// VRManager is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
  // (at your option) any later version.
  // 
  // VRManager is distributed in the hope that it will be useful,
  // but WITHOUT ANY WARRANTY; without even the implied warranty of
  // MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  // GNU General Public License for more details.
  // 
  // You should have received a copy of the GNU Lesser General Public License
  // along with VRManager; if not, write to the Free Software
  // Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  
  
  #include "vrmanager.h"
  #include "xmlHelpers.h"
  #include "debug.h"
  #include <errno.h>
  //#include "dmalloc.h"
  
  // Shared object stuffs
  //  void __attribute__ ((constructor)) my_init(void) {}
  //  void __attribute__ ((destructor)) my_fini(void) {}
  
    // VRManager methods
  
  VRManager::VRManager() {
  }
  
  VRManager::~VRManager() {
      // Delete all objects
    if( trackerSystemHash.size() > 0 ) {
      msg(M_INFO, "VRManager: destroying all trackerSystems..\n");
      for( hash_map<const char*, TrackerSystem*,  hash<const char*>, eqstr>::iterator i = trackerSystemHash.begin(); i != trackerSystemHash.end(); i++ ) {
        msg(M_DEBUG, "  destroying trackerSystem %s\n", (*i).first);
        (*i).second->cleanup();
        delete (*i).second;
      }
    }
    if( surfaceHash.size() > 0 ) {
    msg(M_INFO, "VRManager: destroying all surfaces..\n");
    for( hash_map<const char*, Surface*,  hash<const char*>, eqstr>::iterator i = surfaceHash.begin(); i != surfaceHash.end(); i++ ) {
      msg(M_DEBUG, "  destroying surface %s\n", (*i).first);
      delete (*i).second;
    }
  }
  if( pointerHash.size() > 0 ) {
    msg(M_INFO, "VRManager: destroying all pointers..\n");
    for( hash_map<const char*, Pointer*,  hash<const char*>, eqstr>::iterator i = pointerHash.begin(); i != pointerHash.end(); i++ ) {
      msg(M_DEBUG, "  destroying pointer %s\n", (*i).first);
      delete (*i).second;
    }
  }
  if( viewportHash.size() > 0 ) {
    msg(M_INFO, "VRManager: destroying all viewports..\n");
    for( hash_map<const char*, Viewport*,  hash<const char*>, eqstr>::iterator i = viewportHash.begin(); i != viewportHash.end(); i++ ) {
      msg(M_DEBUG, "  destroying viewport %s\n", (*i).first);
      delete (*i).second;
    }
  }

  // tell poll thread to exit and wait
    
  msg(M_DEBUG, "Waiting for config poll thread to exit..\n");
  pollConfigFiles = false;
  pthread_join(pollThread, NULL);
}


VRManager::VRManager(const char* configFile) {
    // Locate first available config file

    // Check environment variable first
  char* envConfigFile = getenv("VRMANAGER_CONFIG_FILE");
  if( envConfigFile != NULL ) {
    configFile = envConfigFile;
  }
    // Next, see if a file name was passed in.
    // If not, use the default config name
  if( configFile == NULL ) {
    baseConfigFile = new char[strlen(DEFAULT_CONFIG_FILE)+1];
    strcpy(baseConfigFile, DEFAULT_CONFIG_FILE);
  }
  else {
    baseConfigFile = new char[strlen(configFile)+10];
    strcpy(baseConfigFile, configFile);
  }

    // parse
  xmlInitParser();
  updateConfig();

    // Check for config updates every 100ms
  pollInterval = CONF_POLL_INTERVAL;
    // This should be configured someplace..
  pollConfigFiles = true;

    // Start polling for changes to the config file
  if( pollConfigFiles )  {
    startConfPoll();
  }

}


void VRManager::addPollFile(char* filename) {
    // Make sure we haven't already seen this file
  for( slist<fileData*>::iterator i = pollFiles.begin(); i != pollFiles.end(); i++ ) {
    if( strcmp((*i)->name, filename) == 0 ) {
      return;
    }
  }

    // Create new struct to hold file name and modification time
  fileData *data = new fileData;
  data->name = new char[(strlen(filename)+1)];
  sprintf(data->name, "%s", filename);
//  strncpy(data->name, filename, strlen(filename));
  struct stat statBuf;
  stat(filename, &statBuf);
  data->mod_time = statBuf.st_mtime;

    // Add the struct into the list of files to poll
  pollFiles.push_front(data);
}



void VRManager::startConfPoll() {
  pollConfigFiles = true;
  pthread_create( &pollThread, NULL, confPollWrapper, (void*) this);
}


void *VRManager::confPollWrapper(void* vrm) {
  VRManager* vPtr = (VRManager*)vrm;
  vPtr->confPoll();
}


void VRManager::confPoll() {
  struct stat statBuf;
  bool shouldUpdate, foundAllFiles;

  while( pollConfigFiles ) {
    shouldUpdate = false;
    foundAllFiles = true;

    for( slist<fileData*>::iterator i = pollFiles.begin(); i != pollFiles.end(); i++ ) {
      int e, er;
      e = stat((*i)->name, &statBuf);

      if( e != 0 ) {
        er = errno;
        msg(M_WARN, "stat returned %d for file \"%s\": %s  (delaying update..)\n", e, (*i)->name, strerror(er));
        foundAllFiles = false;
      }

      if( (*i)->mod_time != statBuf.st_mtime ) {
        (*i)->mod_time = statBuf.st_mtime;
        shouldUpdate = true;
      }


    }

    if( shouldUpdate && foundAllFiles ) {
      msg(M_INFO, "Detected config change, updating...\n");
      updateConfig();
    }

    usleep(pollInterval);
  }
  pthread_exit(NULL);
}



Matrix VRManager::xmlCsMatrix(xmlNode *node) {
  Matrix cs;
  xmlNode *cur_node = NULL;
  forEachChild(node, cur_node) {
    if( xmlNameEq(cur_node, "translate") ) {
      if( nodeHasProp(cur_node, "vector") ) {
        Vector t = strToVector(xmlGetProp(cur_node, "vector"));
        cs = translateMatrix(Vector() - t) * cs;
      }
      else if( nodeHasProp(cur_node, "pointer") ) {
        Vector t = getPointerLocation(xmlGetProp(cur_node, "pointer"));
        msg(M_DEBUG, "Setting translation (from pointer %s) to %0.2f %0.2f %0.2f\n", xmlGetProp(cur_node, "pointer"), t[0], t[1], t[2]);
        cs = translateMatrix(Vector() - t) * cs;
      }
      else {
        assertNodeHasProp(cur_node, "vector");
      }
    }
    else if( xmlNameEq(cur_node, "scale") ) {
      assertNodeHasProp(cur_node, "vector");
      Vector s = strToVector(xmlGetProp(cur_node, "vector"));
      cs = scaleMatrix(Vector(1.0/s[0], 1.0/s[1], 1.0/s[2])) * cs;
    }
    else if( xmlNameEq(cur_node, "rotate") ) {
      assertNodeHasProp(cur_node, "angle");
      assertNodeHasProp(cur_node, "axis");
      float angle = atof(xmlGetProp(cur_node, "angle"));
      Vector a = strToVector(xmlGetProp(cur_node, "axis"));
      cs = rotateMatrix(a, -angle*D2R) * cs;
    }
    else {
      msg(M_WARN, "[line %d]:Ignoring unknown coordSystem property \"%s\" (options are translate, scale, rotate)\n", xmlGetLineNo(cur_node), (const char*)cur_node->name);
    }
  }
  return cs;
}


void VRManager::updateConfig() {

    // Reset the list of config files to poll
  pollFiles.clear();
  addPollFile(baseConfigFile);

    // Forget about all the surfaces we know
  clearSurfaces();

    // Load the chosen config file
  xmlNode* root_element = getXmlRoot(baseConfigFile);

    // Evaluate Conditional blocks
  evalConditionals(root_element);


    // Check it for include files (included files are also evaluated for conditionals)
  expandIncludes(root_element);


  xmlNode *cur_node = NULL;
  forEachChild(root_element, cur_node) {
    if( xmlNameEq(cur_node, "coordSystem") ) {
        // Clear out old CS first
      coordSystem = Matrix();
        // Read in new one
      coordSystem = xmlCsMatrix( cur_node );
//        invCoordSystem = xmlCsMatrix( cur_node ).inverse();
      invCoordSystem = coordSystem.inverse();
    }
    else if( xmlNameEq(cur_node, "trackerSystem") ) {
      parseTrackerSystem(cur_node);
    }
    else if( xmlNameEq(cur_node, "pointer") ) {
      parsePointer(cur_node);
    }
    else if( xmlNameEq(cur_node, "viewport") ) {
      parseViewport(cur_node);
    }
    else if( xmlNameEq(cur_node, "surface") ) {
      parseSurface(cur_node);
    }
    else {
      msg(M_WARN, "Tag \"%s\" ignored. (options are \"coordSystem\" \"trackerSystem\" \"pointer\" \"viewport\")\n", xmlNodeName(cur_node));
    }
  }
  
    // Update the surface information
  updateSurfaces();

  msg(M_DEBUG, "Final result: \n%s\n", xmlTreeString(root_element));
}


void VRManager::clearSurfaces() {
    // This might be a memory leak?
  for( hash_map<const char*, Surface*,  hash<const char*>, eqstr>::iterator i = surfaceHash.begin(); i != surfaceHash.end(); i++ ) {
//  for( map<const char*, Surface*,  eqstr>::iterator i = surfaceHash.begin(); i != surfaceHash.end(); i++ ) {
    delete (*i).second;
  }
  surfaceHash.clear();
}


  // Make sure every viewport has a Surface associated with it
  // Viewports with similar surfaces should automatically share the same Surface

void VRManager::updateSurfaces() {

    // Iterate over all viewports, make sure each has a surface.
  for( hash_map<const char*, Viewport*,  hash<const char*>, eqstr>::iterator v = viewportHash.begin(); v != viewportHash.end(); v++ ) {
    Viewport* vp = (*v).second;
    if( vp->getSurface() == NULL ) {
        // Find corners
      Pointer* a = vp->getCorner("ul");
      Pointer* b = vp->getCorner("ur");
      Pointer* c = vp->getCorner("br");

        // Make a new surface
      char* name = new char[strlen(vp->getVrmName())+9];
      sprintf(name, "%s_surface", vp->getVrmName());

      Surface* s = new Surface_3Point(a, b, c, this, name);
      bool new_surface = true;

        // Check existing surfaces for similarity
      for( hash_map<const char*, Surface*,  hash<const char*>, eqstr>::iterator i = surfaceHash.begin(); i != surfaceHash.end(); i++ ) {
        if( s->isSimilar((*i).second) ) {
          delete s;
          s = (*i).second;
          new_surface = false;
          break;
        }
      }

      if( new_surface ) {
          // Inform VRManager of the new surface
        surfaceHash[name] = s;
      }

      vp->setSurface(s);
      s->addViewport(vp);
    }
  }
}


  // Search recursively for include tags and replace them
  // with the content of their src="" file

void VRManager::expandIncludes(xmlNode* root) {
  xmlNode *cur_node, *next_node;
  for (cur_node = root->children; cur_node; cur_node = next_node) {
    next_node = findElementNode(cur_node->next);

      // If this is an include tag, parse the file and insert its toplevel nodes in place.
    if( xmlNameEq(cur_node, "include") ) {
      xmlNode* old_node = cur_node;
      assertNodeHasProp(cur_node, "src");
      char* file = xmlGetProp(cur_node, "src");

        // Load the included file
      addPollFile(file);
      xmlNode* newRoot = getXmlRoot(file);

      evalConditionals(newRoot);

        // Insert the new XML and reset the pointer
        // so that we end up parsing all of the XML
        // we just included
      next_node = replaceNodeWithChildren(cur_node, newRoot);
    }
    else {
      expandIncludes(cur_node);
    }
  }
}


void VRManager::evalConditionals(xmlNode *root) {
  xmlNode *cur_node;
  xmlNode *next_cur_node;
  bool madeChanges = true;
  while( madeChanges ) {
    madeChanges = false;
    for (cur_node = root->children; cur_node; cur_node = next_cur_node) {
      next_cur_node = findElementNode(cur_node->next);

        // If this is a <select> tag, evaluate the condition and replace with the correct response
      if( xmlNameEq(cur_node, "select") ) {
        char* match = NULL;
        if( xmlPropEq(cur_node, "type", "host") ) {
          match  = new char[512];
          gethostname(match, 512);
        }
        else if( xmlPropEq(cur_node, "type", "user") ) {
          match = strdup(cuserid(NULL));
        }
        else if( xmlPropEq(cur_node, "type", "command") ) {
          char cmd[512];
          
          char procfile[128];
          sprintf(procfile, "/proc/%d/cmdline", getpid());
          FILE* fp = fopen(procfile, "r");
          fscanf(fp, "%511s", cmd);
          fclose(fp);
          char *buf, *buf1;
          buf = strtok(cmd, "/");
          buf1 = buf;
          while( (buf = strtok(NULL, "/")) != NULL ) {
            buf1 = buf;
          }
          match = strdup(buf1);
        }
        else if( xmlPropEq(cur_node, "type", "environment") ) {
          char* env = getenv(xmlGetProp(cur_node, "name"));
          if( env != NULL ) {
            match = strdup(env);
          }
        }
        else {
          msg(M_ERR, "Select type \"%s\" is not supported. Use one of: \"host\" \"user\" \"command\" \"environment\"\n", xmlGetProp(cur_node, "type"));
          return;
        }
        if( match == NULL ) {
          match = new char[2];
          strncpy(match, "", 1);
        }
        
        xmlNode *case_node;
        xmlNode *correct_node = NULL;
        xmlNode *default_node = NULL;
        forEachChild(cur_node, case_node) {
          if (xmlNameEq(case_node, "case")) {
            if( xmlPropEq(case_node, "value", match) ) {
              correct_node = case_node;
            }
            else if( xmlPropEq(case_node, "value", "*") ) {
              default_node = case_node;
            }
          }
        }

        if( correct_node != NULL ) {
          replaceNodeWithChildren(cur_node, correct_node);
          madeChanges = true;
        }
        else if( default_node != NULL ) {
          replaceNodeWithChildren(cur_node, default_node);
          madeChanges = true;
        }
        else {
          msg(M_WARN, "No case matched and no default case (*) on line %d\n", xmlGetLineNo(cur_node));
        }
        
        if( match != NULL ) {
          delete[] match;
        }

      }
      else {
        evalConditionals(cur_node);
      }
    }  // end for
  } // end while
}








void VRManager::setButtonCallback(void (*callbackFn)(int, int, void*), void* userData) {
  hash_map<const char*, TrackerSystem*, hash<const char*>, eqstr>::iterator i = trackerSystemHash.begin();
  hash_map<const char*, TrackerSystem*, hash<const char*>, eqstr>::iterator e = trackerSystemHash.end();
//  map<const char*, TrackerSystem*, eqstr>::iterator i = trackerSystemHash.begin();
//  map<const char*, TrackerSystem*, eqstr>::iterator e = trackerSystemHash.end();
  for( int n=0; n<trackerSystemHash.size(); n++ ) {
    i->second->setButtonCallback(callbackFn, userData);
    if( i == e )
      break;
    i++;
  }
}





Viewport* VRManager::getViewport(const char* name) {
  if( viewportHash.size() > 0 && name != NULL ) {
    hash_map<const char*, Viewport*, hash<const char*>, eqstr>::iterator i = viewportHash.find(name);
    if( i != viewportHash.end() ) {
      return (*i).second;
    }
    else {
      return NULL;
    }
//  return viewportHash[name];
  }
  else {
    return NULL;
  }
}

Pointer* VRManager::getPointer(const char* name) {
  char* n = strdup(name);
  if( name != NULL ) {
    Pointer* p = NULL;
    //p = pointerHash[name];
    if( pointerHash.size() > 0 ) {
      hash_map<const char*, Pointer*, hash<const char*>, eqstr>::iterator i = pointerHash.find(name);
      if( i != pointerHash.end() ) {
        p = (*i).second;
      }
    }
    
    if( p == NULL ) {
      char* vp_name = strtok(n, ".");
      Viewport* vp = getViewport(vp_name);
      char* c_name = strtok(NULL, "");
      if( c_name != NULL ) {
        msg(M_INFO, "Checking for pointer \"%s\" in viewport \"%s\"\n", c_name, vp_name);
        if( vp != NULL ) {
          p = vp->getCorner(c_name);
        }
      }
    }
    return p;
  }
  else {
    return NULL;
  }
}

TrackerSystem* VRManager::getTrackerSystem(const char *name) {
  if( name != NULL && trackerSystemHash.size() > 0 ) {
    hash_map<const char*, TrackerSystem*, hash<const char*>, eqstr>::iterator i = trackerSystemHash.find(name);
    if( i != trackerSystemHash.end() ) {
      return (*i).second;
    }
    else {
      return NULL;
    }
//    return trackerSystemHash[name];
  }
  else {
    return NULL;
  }
}


void VRManager::addPointer(const char* name, Pointer* p) {
  if( name != NULL )
    pointerHash[strdup(name)] = p;
}

void VRManager::addViewport(const char* name, Viewport* v) {
  
  if( name != NULL )
    viewportHash[strdup(name)] = v;
}



Pointer* VRManager::parsePointer(xmlNode *node) {
  msg(M_INFO, "== Parsing pointer %s\n", xmlGetProp(node, "name"));
    
  // All pointers must have names
  assertNodeHasProp(node, "name");
  char* name = xmlGetProp(node, "name");
  Pointer* ret = getPointer(name);
  if( ret == NULL ) {
    ret = createNewPointer(node, this, name);
    addPointer(name, ret);
  }
  else {
    ret->reconfigure(node);
  }
  return ret;
}


Viewport* VRManager::parseViewport(xmlNode *node) {
  msg(M_INFO, "== Parsing viewport %s\n", xmlGetProp(node, "name"));
  assertNodeHasProp(node, "name");
  char* name = xmlGetProp(node, "name");
  Viewport* ret = getViewport(name);
  
  if( ret == NULL ) {
    ret = createNewViewport(node, this, name);
    addViewport(name, ret);
  }
  else {
    ret->reconfigure(node, this);
  }
  return ret;
}


void VRManager::addTrackerSystem(const char* name, TrackerSystem* trackerSystem) {
  trackerSystemHash[strdup(name)] = trackerSystem;
}


TrackerSystem* VRManager::parseTrackerSystem(xmlNode *node) {
  msg(M_INFO, "== Parsing trackerSystem %s\n", xmlGetProp(node, "name"));
  assertNodeHasProp(node, "name");
  char* trackerName = xmlGetProp(node, "name");
  TrackerSystem* ret = getTrackerSystem(trackerName);
  
  if( ret == NULL ) {
    ret = createNewTrackerSystem(node, this);
    addTrackerSystem(trackerName, ret);
  }
  else {
    ret->reconfigure(node, this);
  }
  return ret;
}



int* VRManager::getViewportDrawArea(char* vp) {
  Viewport* v = getViewport(vp);
  if( v == NULL ) {
    msg(M_ERR, "Viewport \"%s\" does not exist!\n", vp);
    return NULL;
  }
  return v->getDrawArea();
}

Matrix VRManager::getProjectionMatrix(const char* viewport, const char* pointer, float near, float far) {
  Pointer* p = getPointer(pointer);
  Viewport* v = getViewport(viewport);
  if( p == NULL ) {
    msg(M_ERR, "Pointer \"%s\" does not exist!\n", pointer);
    return Matrix();
  }
  if( v == NULL ) {
    msg(M_ERR, "Viewport \"%s\" does not exist!\n", viewport);
    return Matrix();
  }
  return v->generateProjectionMatrix(p, near, far) * invCoordSystem;
}

float* VRManager::getGlProjectionMatrix(const char* viewport, const char* pointer, float near, float far) {
  Pointer* p = getPointer(pointer);
  Viewport* v = getViewport(viewport);
  if( p == NULL ) {
    msg(M_ERR, "Pointer \"%s\" does not exist!\n", pointer);
    return 0;
  }
  if( v == NULL ) {
    msg(M_ERR, "Viewport \"%s\" does not exist!\n", viewport);
    return 0;
  }
  return (v->generateProjectionMatrix(p, near, far) * invCoordSystem).glMatrix();
}

float* VRManager::getGlOrthoMatrix(const char* viewport, float near, float far) {
  Viewport* v = getViewport(viewport);
  if( v == NULL ) {
    msg(M_ERR, "Viewport \"%s\" does not exist!\n", viewport);
    return 0;
  }
  return (v->generateOrthoMatrix(near, far) * invCoordSystem).glMatrix();
}


Vector VRManager::getPointerLocation(const char* ptr) {
  Pointer* p = getPointer(ptr);
  if( p == NULL ) {
    msg(M_ERR, "Pointer \"%s\" does not exist!\n", ptr);
  }
  return getPointerLocation(p);
}


Matrix VRManager::getPointerTransform(const char* ptr) {
  Pointer* p = getPointer(ptr);
  if( p == NULL ) {
    msg(M_ERR, "Pointer \"%s\" does not exist!\n", ptr);
  }
  return getPointerTransform(p);
}

Matrix VRManager::getPointerInvTransform(const char* ptr) {
  Pointer* p = getPointer(ptr);
  if( p == NULL ) {
    msg(M_ERR, "Pointer \"%s\" does not exist!\n", ptr);
  }
  return getPointerInvTransform(p);
}

Vector VRManager::getPointerLocation(Pointer* p) {
  if( p == NULL ) {
    return Vector();
  }
  return coordSystem * p->getLocation();
}


Matrix VRManager::getPointerTransform(Pointer* p) {
  if( p == NULL ) {
    return Matrix();
  }
  return coordSystem * p->getTransform();
}

Matrix VRManager::getPointerInvTransform(Pointer* p) {
  if( p == NULL ) {
    return Matrix();
  }
  return coordSystem * p->getInvTransform();
}

void VRManager::adjustBuffer(char* vp, int x, int y, int w, int h) {
  Viewport* v = getViewport(vp);
  v->correctDistortion(x, y, w, h);
    // Overlay blending stencil
  v->drawMask();
}




void VRManager::addSurface(const char* name, Surface* s) {
  if( name != NULL )
    surfaceHash[strdup(name)] = s;
}

Surface* VRManager::getSurface(const char* name) {
  if( name != NULL ){
    return surfaceHash[name];
  }
  else {
    msg(M_ERR, "Surface \"name\" does not exist.\n");
    return NULL;
  }
}

Surface* VRManager::parseSurface(xmlNode *node) {
  msg(M_INFO, "== Parsing surface %s\n", xmlGetProp(node, "name"));
  assertNodeHasProp(node, "name");
  char* name = xmlGetProp(node, "name");
  Surface* ret = getSurface(name);

  if( ret == NULL ) {
    ret = createNewSurface(node, this);
    addSurface(name, ret);
  }
  else {
    ret->reconfigure(node);
  }
  return ret;
}
