// Copyright 2004,2005 Luke Campagnola <lcampagn@mines.edu>, Colorado School of Mines
// 
// This file is part of VRManager.
// 
// VRManager is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
// 
// VRManager is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with VRManager; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#ifndef VRMANAGER_H
#define VRMANAGER_H

#include "vectormath.h"
#include <pthread.h>  // threads used to poll for config file updates
#include <string.h>
#include <sys/stat.h>
#include "Pointer.h"
#include "Tracker.h"
#include "TrackerSystem.h"
#include "Viewport.h"
#include "Surface.h"
#include "xmlHelpers.h"

#define DEFAULT_CONFIG_FILE "/etc/vrmanagerConfig.xml"
#define CONF_POLL_INTERVAL 1000000   // Time interval for polling the config file (in us)


  // This class manages a full system of pointers, viewports, and trackers
class VRManager {
  private:

    struct fileData {
      char* name;
      time_t mod_time;
    };
  
  
    Matrix coordSystem, invCoordSystem;   // Transforms from internal to external coordinate systems.
  
    
      // Data
/*    map<const char*, Pointer*,  eqstr> pointerHash;
    map<const char*, Viewport*, eqstr> viewportHash;
    map<const char*, TrackerSystem*, eqstr> trackerSystemHash;
    map<const char*, Surface*,  eqstr> surfaceHash;*/
    hash_map<const char*, Pointer*,  hash<const char*>, eqstr> pointerHash;
    hash_map<const char*, Viewport*, hash<const char*>, eqstr> viewportHash;
    hash_map<const char*, TrackerSystem*, hash<const char*>, eqstr> trackerSystemHash;
    hash_map<const char*, Surface*,  hash<const char*>, eqstr> surfaceHash;
void checkHash();    
    
      // Configuration files and polling
    char* baseConfigFile;
    slist<fileData*> pollFiles;
    bool pollConfigFiles;
    int pollInterval;
    pthread_t pollThread;
    

    void addPollFile(char* filename);    
    void startConfPoll();
    static void *confPollWrapper(void* vrm);
    void confPoll();
    void updateConfig();

    void clearSurfaces();
    void updateSurfaces();
    
    void expandIncludes(xmlNode *root);
    void evalConditionals(xmlNode *root);
    
    void addPointer(const char* name, Pointer* pointer);

    void addViewport(const char* name, Viewport* viewport);
    
    void addTrackerSystem(const char* name, TrackerSystem* trackerSystem);

    void addSurface(const char* name, Surface* surface);
  
  public:
    
    Matrix xmlCsMatrix(xmlNode *node);
    Surface* parseSurface(xmlNode *node);
    TrackerSystem* parseTrackerSystem(xmlNode *node);
    Pointer* parsePointer(xmlNode *node);
    Viewport* parseViewport(xmlNode *node);
    
    ~VRManager();
    VRManager(const char* configFile);
    VRManager();
    
    TrackerSystem* getTrackerSystem(const char *name);
    Viewport* getViewport(const char* name);
    Pointer* getPointer(const char* name);
    Surface* getSurface(const char* name);
    
            
      // These are about the only functions interesting to the end programmer    
    Vector getPointerLocation(const char* ptr);
    Matrix getPointerTransform(const char* ptr);
    Matrix getPointerInvTransform(const char* ptr);
    Vector getPointerLocation(Pointer* ptr);
    Matrix getPointerTransform(Pointer* ptr);
    Matrix getPointerInvTransform(Pointer* ptr);
    Matrix getProjectionMatrix(const char* viewport, const char* pointer, float near, float far);   // Returns the Matrix needed to correctly render a perspective viewport
    float* getGlProjectionMatrix(const char* viewport, const char* pointer, float near, float far);   // Returns the Matrix needed to correctly render a perspective viewport
    float* getGlOrthoMatrix(const char* viewport, float near, float far);
    void setButtonCallback(void (*callbackFn)(int, int, void*), void* userData=NULL);
    int* getViewportDrawArea(char* vp);
    
    void adjustBuffer(char* vp, int x, int y, int w, int h);
    
    friend class Viewport;
    friend class Surface;
    friend class Pointer;
};



#endif
