// Copyright 2004,2005 Luke Campagnola <lcampagn@mines.edu>, Colorado School of Mines
// 
// This file is part of VRManager.
// 
// VRManager is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
// 
// VRManager is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with VRManager; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "xmlHelpers.h"
#include "debug.h"


  // Iterates over an xmlNode's siblings, looking for the first that has type = XML_ELEMENT_NODE
xmlNode* findElementNode(xmlNode* node) {
  while( node != NULL && node->type != XML_ELEMENT_NODE ) {
    node = node->next;
  }
  return node;
}


  // Replaces one node (dst) with the children of another node (src), returning the first child node.

xmlNode* replaceNodeWithChildren(xmlNode* dst, xmlNode* src) {
  xmlNode *firstChild, *child, *ret, *next, *add_node;
  firstChild = NULL;
  add_node = dst;
  
  for (child = src->children; child; child = next) {
    next = findElementNode(child->next);
    if( firstChild == NULL ) {
      firstChild = child;
    }
    xmlNode* new_node = xmlAddNextSibling(add_node, child);
    add_node = child;
  }        
  
  xmlUnlinkNode(dst);
//  xmlFreeNode(dst);
  return firstChild;
}

xmlNode* getXmlRoot(char* filename) {
  if( filename == NULL ) {
    msg(M_WARN, "Can not parse file; null string passed as filename.\n");
  }
  if( strlen(filename) < 1 ) {
    msg(M_WARN, "Can not parse file; blank string passed as filename.\n");
  }
    // Make sure file exists and is readable.
  struct stat fs;
  stat(filename, &fs);
  if( ! (fs.st_mode & S_IFREG) ) {
    msg(M_WARN, "File \"%s\" does not exist or is not a file.\n", filename);
  }
      
  xmlDocPtr doc;


    // Don't use libxml dictionaries. They probably speed things up, 
    // but the implementation is broken, imo.
  msg(M_INFO, "Reading file \"%s\"\n", filename);
  doc = xmlReadFile(filename, NULL, XML_PARSE_NODICT|XML_PARSE_PEDANTIC);
//  doc = xmlReadFile(filename, NULL, 0);

  if (doc == NULL) {
    msg(M_WARN, "Failed to parse %s\n", filename);
    return NULL;
  }
    
//  xmlFreeDoc(doc);
//  xmlCleanupParser();

  return xmlDocGetRootElement(doc);
}


char* xmlTreeString(xmlNode *node, char* indent) {
  char* str = new char[32768];
  char* buf = new char[1024];
  sprintf(str, "%s<%s", indent, (char*)node->name);
  
  for(xmlAttr *p = node->properties; p; p = p->next) {
    sprintf(buf, " %s=\"%s\"", (char*)p->name, (char*)xmlGetProp(node, p->name));
    strcat(str, buf);
  }
  
  if( node->children == NULL ) {
    strcat(str, "/>\n"); 
  }
  else {
    strcat(str, ">\n"); 
    
    char* new_indent = new char[512];
    sprintf(new_indent, "  %s", indent);
    
    
    xmlNode* cur_node;
    forEachChild(node, cur_node) {
      char* nodeStr = xmlTreeString(cur_node, new_indent);
      strcat(str, nodeStr);
      delete[] nodeStr;
    }
    delete[] new_indent;
    
    char* closeTag = new char[512];
    sprintf(closeTag, "%s</%s>\n", indent, (char*)node->name);
    strcat(str, closeTag);
    delete[] closeTag;
  }
  delete buf;
  return str;
}


char* xmlNodeString(xmlNode *node) {
  int slen, blen;
  slen = strlen((char*)node->name) + 3;
  blen = 0;
  for(xmlAttr *p = node->properties; p; p = p->next) {
    int l = strlen((char*)p->name) + strlen((char*)xmlGetProp(node, p->name)) + 4;
    slen += l;
    blen = (blen>l ? blen : l);
  }
  blen += 1;
  
  char* str = new char[slen];
  char* buf = new char[blen];
  sprintf(str, "<%s", (char*)node->name);

  for(xmlAttr *p = node->properties; p; p = p->next) {
    sprintf(buf, " %s=\"%s\"", (char*)p->name, (char*)xmlGetProp(node, p->name));
    strcat(str, buf);
  }
  strcat(str, ">");
  delete buf;
  return str;
}


char* xmlGetProp(xmlNode *node, const char* property) {
  return (char*)xmlGetProp(node, (const xmlChar*)property);
}

void xmlSetProp(xmlNode *node, const char* property, const char* value) {
  xmlSetProp(node, (const xmlChar*)property, (const xmlChar*)value);
}

void xmlUnsetProp(xmlNode *node, const char* property) {
  xmlUnsetProp(node, (const xmlChar*)property);
}


bool xmlPropEq(xmlNode *node, const char* property, const char* value) {
  int c = strcmp(xmlGetProp(node, property), value);
  if( c == 0 ) {
    return true;
  }
  else {
    return false;
  }
}

bool xmlNameEq(xmlNode *node, const char* value) {
  int c = strcmp((const char*)node->name, value);
  if( c == 0 ) {
    return true;
  }
  else {
    return false;
  }
}

char* xmlNodeName(xmlNode* node) {
  return (char*)node->name;
}

Vector strToVector(char* str) {
  float x, y, z;
  x = atof(strtok(str, ","));
  y = atof(strtok(NULL, ","));
  z = atof(strtok(NULL, ",\0"));
  return(Vector(x, y, z));
}

void assertNodeHasProp(xmlNode *node, char *attr) {
  if( xmlGetProp(node, attr) == NULL ) {
    msg(M_ERR, "Parse error [line %d]: \"%s\" must have attribute \"%s\".\n", xmlGetLineNo(node), (char*)node->name, attr);
    exit(1);
  }
}

bool nodeHasProp(xmlNode *node, char *attr) {
  if( xmlGetProp(node, attr) == NULL ) {
    return(false);
  }
  else {
    return(true);
  }
}
  

bool nodeHasChildren(xmlNode *node) {
  return (node->children != NULL);
}

void setChildrenProp(xmlNode* node, char* name, char* prop, char* value) {
  xmlNode* n;
  forEachChild(node, n) {
    if (name == NULL || xmlNameEq(n, name)) {
      xmlSetProp(n, prop, value);
    }
  }
}

void moveChildrenAfterNode(xmlNode* src, xmlNode* dst) {
  xmlNode* ptr = dst;
  xmlNode* cur, *next;
  for( cur = src->children; cur != NULL; cur = next) {
    next = findElementNode(cur->next);
    xmlAddNextSibling(ptr, cur);
    ptr = cur;
  }
}
